# -*- coding: UTF-8 -*-
#    Pluma External Tools plugin
#    Copyright (C) 2005-2006  Steve Frécinaux <steve@istique.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

__all__ = ('ExternalToolsPlugin', 'ExternalToolsWindowHelper',
           'Manager', 'OutputPanel', 'Capture', 'UniqueById')

import pluma
import gtk
from manager import Manager
from library import ToolLibrary
from outputpanel import OutputPanel
from capture import Capture
from functions import *

class ToolMenu(object):
    ACTION_HANDLER_DATA_KEY = "ExternalToolActionHandlerData"
    ACTION_ITEM_DATA_KEY = "ExternalToolActionItemData"

    def __init__(self, library, window, menupath):
        super(ToolMenu, self).__init__()
        self._library = library
        self._window = window
        self._menupath = menupath

        self._merge_id = 0
        self._action_group = gtk.ActionGroup("ExternalToolsPluginToolActions")
        self._signals = []

        self.update()

    def deactivate(self):
        self.remove()

    def remove(self):
        if self._merge_id != 0:
            self._window.get_ui_manager().remove_ui(self._merge_id)
            self._window.get_ui_manager().remove_action_group(self._action_group)
            self._merge_id = 0

        for action in self._action_group.list_actions():
            handler = action.get_data(self.ACTION_HANDLER_DATA_KEY)

            if handler is not None:
                action.disconnect(handler)

            action.set_data(self.ACTION_ITEM_DATA_KEY, None)
            action.set_data(self.ACTION_HANDLER_DATA_KEY, None)

            self._action_group.remove_action(action)
        
        accelmap = gtk.accel_map_get()

        for s in self._signals:
            accelmap.disconnect(s)
        
        self._signals = []

    def _insert_directory(self, directory, path):
        manager = self._window.get_ui_manager()

        for item in directory.subdirs:
            action_name = 'ExternalToolDirectory%X' % id(item)
            action = gtk.Action(action_name, item.name.replace('_', '__'), None, None)
            self._action_group.add_action(action)

            manager.add_ui(self._merge_id, path,
                           action_name, action_name,
                           gtk.UI_MANAGER_MENU, False)
                           
            self._insert_directory(item, path + '/' + action_name)

        for item in directory.tools:
            action_name = 'ExternalToolTool%X' % id(item)
            action = gtk.Action(action_name, item.name.replace('_', '__'), item.comment, None)
            handler = action.connect("activate", capture_menu_action, self._window, item)

            action.set_data(self.ACTION_ITEM_DATA_KEY, item)
            action.set_data(self.ACTION_HANDLER_DATA_KEY, handler)
            
            # Make sure to replace accel
            accelpath = '<Actions>/ExternalToolsPluginToolActions/%s' % (action_name, )
            
            if item.shortcut:
                key, mod = gtk.accelerator_parse(item.shortcut)
                gtk.accel_map_change_entry(accelpath, key, mod, True)
                
                self._signals.append(gtk.accel_map_get().connect('changed::%s' % (accelpath,), self.on_accelmap_changed, item))
                
            self._action_group.add_action_with_accel(action, item.shortcut)

            manager.add_ui(self._merge_id, path,
                           action_name, action_name,
                           gtk.UI_MANAGER_MENUITEM, False)

    def on_accelmap_changed(self, accelmap, path, key, mod, tool):
        tool.shortcut = gtk.accelerator_name(key, mod)
        tool.save()
        
        self._window.get_data("ExternalToolsPluginWindowData").update_manager(tool)

    def update(self):
        self.remove()
        self._merge_id = self._window.get_ui_manager().new_merge_id()
        self._insert_directory(self._library.tree, self._menupath)
        self._window.get_ui_manager().insert_action_group(self._action_group, -1)
        self.filter(self._window.get_active_document())

    def filter_language(self, language, item):
        if not item.languages:
            return True
        
        if not language and 'plain' in item.languages:
            return True
        
        if language and (language.get_id() in item.languages):
            return True
        else:
            return False

    def filter(self, document):
        if document is None:
            return

        titled = document.get_uri() is not None
        remote = not document.is_local()

        states = {
            'all' : True,
            'local': titled and not remote,
            'remote': titled and remote,
            'titled': titled,
            'untitled': not titled,
        }
        
        language = document.get_language()

        for action in self._action_group.list_actions():
            item = action.get_data(self.ACTION_ITEM_DATA_KEY)

            if item is not None:
                action.set_visible(states[item.applicability] and self.filter_language(language, item))

class ExternalToolsWindowHelper(object):
    def __init__(self, plugin, window):
        super(ExternalToolsWindowHelper, self).__init__()

        self._window = window
        self._plugin = plugin
        self._library = ToolLibrary()

        manager = window.get_ui_manager()

        self._action_group = gtk.ActionGroup('ExternalToolsPluginActions')
        self._action_group.set_translation_domain('pluma')
        self._action_group.add_actions([('ExternalToolManager',
                                         None,
                                         _('Manage _External Tools...'),
                                         None,
                                         _("Opens the External Tools Manager"),
                                         lambda action: plugin.open_dialog()),
                                        ('ExternalTools',
                                        None,
                                        _('External _Tools'),
                                        None,
                                        _("External tools"),
                                        None)])
        manager.insert_action_group(self._action_group, -1)

        ui_string = """
            <ui>
              <menubar name="MenuBar">
                <menu name="ToolsMenu" action="Tools">
                  <placeholder name="ToolsOps_4">
                    <separator/>
                    <menu name="ExternalToolsMenu" action="ExternalTools">
                        <placeholder name="ExternalToolPlaceholder"/>
                    </menu>
                    <separator/>
                  </placeholder>
                  <placeholder name="ToolsOps_5">
                    <menuitem name="ExternalToolManager" action="ExternalToolManager"/>
                  </placeholder>
                </menu>
              </menubar>
            </ui>"""

        self._merge_id = manager.add_ui_from_string(ui_string)

        self.menu = ToolMenu(self._library, self._window,
                             "/MenuBar/ToolsMenu/ToolsOps_4/ExternalToolsMenu/ExternalToolPlaceholder")
        manager.ensure_update()

        # Create output console
        self._output_buffer = OutputPanel(self._plugin.get_data_dir(), window)
        bottom = window.get_bottom_panel()
        bottom.add_item(self._output_buffer.panel,
                        _("Shell Output"),
                        gtk.STOCK_EXECUTE)

    def update_ui(self):
        self.menu.filter(self._window.get_active_document())
        self._window.get_ui_manager().ensure_update()

    def deactivate(self):
        manager = self._window.get_ui_manager()
        self.menu.deactivate()
        manager.remove_ui(self._merge_id)
        manager.remove_action_group(self._action_group)
        manager.ensure_update()

        bottom = self._window.get_bottom_panel()
        bottom.remove_item(self._output_buffer.panel)

    def update_manager(self, tool):
        self._plugin.update_manager(tool)

class ExternalToolsPlugin(pluma.Plugin):
    WINDOW_DATA_KEY = "ExternalToolsPluginWindowData"

    def __init__(self):
        super(ExternalToolsPlugin, self).__init__()
        
        self._manager = None
        self._manager_default_size = None

        ToolLibrary().set_locations(os.path.join(self.get_data_dir(), 'tools'))

    def activate(self, window):
        helper = ExternalToolsWindowHelper(self, window)
        window.set_data(self.WINDOW_DATA_KEY, helper)

    def deactivate(self, window):
        window.get_data(self.WINDOW_DATA_KEY).deactivate()
        window.set_data(self.WINDOW_DATA_KEY, None)

    def update_ui(self, window):
        window.get_data(self.WINDOW_DATA_KEY).update_ui()

    def create_configure_dialog(self):
        return self.open_dialog()

    def open_dialog(self):
        if not self._manager:
            self._manager = Manager(self.get_data_dir())

            if self._manager_default_size:
                self._manager.dialog.set_default_size(*self._manager_default_size)

            self._manager.dialog.connect('destroy', self.on_manager_destroy)

        window = pluma.app_get_default().get_active_window()
        self._manager.run(window)

        return self._manager.dialog

    def update_manager(self, tool):
        if not self._manager:
            return

        self._manager.tool_changed(tool, True)

    def on_manager_destroy(self, dialog):
        self._manager_default_size = [dialog.allocation.width, dialog.allocation.height]
        self._manager = None

# ex:ts=4:et:



# -*- coding: utf-8 -*-
#    Pluma External Tools plugin
#    Copyright (C) 2005-2006  Steve Frécinaux <steve@istique.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

__all__ = ('Capture', )

import os, sys, signal
import locale
import subprocess
import gobject
import fcntl
import glib

class Capture(gobject.GObject):
    CAPTURE_STDOUT = 0x01
    CAPTURE_STDERR = 0x02
    CAPTURE_BOTH   = 0x03
    CAPTURE_NEEDS_SHELL = 0x04
    
    WRITE_BUFFER_SIZE = 0x4000

    __gsignals__ = {
        'stdout-line'  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_STRING,)),
        'stderr-line'  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_STRING,)),
        'begin-execute': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, tuple()),
        'end-execute'  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_INT,))
    }

    def __init__(self, command, cwd = None, env = {}):
        gobject.GObject.__init__(self)
        self.pipe = None
        self.env = env
        self.cwd = cwd
        self.flags = self.CAPTURE_BOTH | self.CAPTURE_NEEDS_SHELL
        self.command = command
        self.input_text = None

    def set_env(self, **values):
        self.env.update(**values)

    def set_command(self, command):
        self.command = command

    def set_flags(self, flags):
        self.flags = flags

    def set_input(self, text):
        self.input_text = text

    def set_cwd(self, cwd):
        self.cwd = cwd

    def execute(self):
        if self.command is None:
            return

        # Initialize pipe
        popen_args = {
            'cwd'  : self.cwd,
            'shell': self.flags & self.CAPTURE_NEEDS_SHELL,
            'env'  : self.env
        }
        
        if self.input_text is not None:
            popen_args['stdin'] = subprocess.PIPE
        if self.flags & self.CAPTURE_STDOUT:
            popen_args['stdout'] = subprocess.PIPE
        if self.flags & self.CAPTURE_STDERR:
            popen_args['stderr'] = subprocess.PIPE

        self.tried_killing = False
        self.idle_write_id = 0
        self.read_buffer = ''
        
        try:
            self.pipe = subprocess.Popen(self.command, **popen_args)
        except OSError, e:
            self.pipe = None
            self.emit('stderr-line', _('Could not execute command: %s') % (e, ))
            return
        
        # Signal
        self.emit('begin-execute')
        
        if self.flags & self.CAPTURE_STDOUT:
            # Set non blocking
            flags = fcntl.fcntl(self.pipe.stdout.fileno(), fcntl.F_GETFL) | os.O_NONBLOCK
            fcntl.fcntl(self.pipe.stdout.fileno(), fcntl.F_SETFL, flags)

            gobject.io_add_watch(self.pipe.stdout,
                                 gobject.IO_IN | gobject.IO_HUP,
                                 self.on_output)

        if self.flags & self.CAPTURE_STDERR:
            # Set non blocking
            flags = fcntl.fcntl(self.pipe.stderr.fileno(), fcntl.F_GETFL) | os.O_NONBLOCK
            fcntl.fcntl(self.pipe.stderr.fileno(), fcntl.F_SETFL, flags)

            gobject.io_add_watch(self.pipe.stderr,
                                 gobject.IO_IN | gobject.IO_HUP,
                                 self.on_output)

        # IO
        if self.input_text is not None:
            # Write async, in chunks of something
            self.write_buffer = str(self.input_text)

            if self.idle_write_chunk():
                self.idle_write_id = gobject.idle_add(self.idle_write_chunk)

        # Wait for the process to complete
        gobject.child_watch_add(self.pipe.pid, self.on_child_end)

    def idle_write_chunk(self):
        if not self.pipe:
            self.idle_write_id = 0
            return False

        try:
            l = len(self.write_buffer)
            m = min(l, self.WRITE_BUFFER_SIZE)
         
            self.pipe.stdin.write(self.write_buffer[:m])
            
            if m == l:
                self.write_buffer = ''
                self.pipe.stdin.close()
                
                self.idle_write_id = 0

                return False
            else:
                self.write_buffer = self.write_buffer[m:]
                return True
        except IOError:
            self.pipe.stdin.close()
            self.idle_write_id = 0

            return False

    def on_output(self, source, condition):
        if condition & (glib.IO_IN | glib.IO_PRI):
            line = source.read()

            if len(line) > 0:
                try:
                    line = unicode(line, 'utf-8')
                except:
                    line = unicode(line,
                                   locale.getdefaultlocale()[1],
                                   'replace')

                self.read_buffer += line
                lines = self.read_buffer.splitlines(True)
                
                if not lines[-1].endswith("\n"):
                    self.read_buffer = lines[-1]
                    lines = lines[0:-1]
                else:
                    self.read_buffer = ''

                for line in lines:
                    if not self.pipe or source == self.pipe.stdout:
                        self.emit('stdout-line', line)
                    else:
                        self.emit('stderr-line', line)

        if condition & ~(glib.IO_IN | glib.IO_PRI):
            if self.read_buffer:
                if source == self.pipe.stdout:
                    self.emit('stdout-line', self.read_buffer)
                else:
                    self.emit('stderr-line', self.read_buffer)

                self.read_buffer = ''

            self.pipe = None

            return False
        else:
            return True

    def stop(self, error_code = -1):
        if self.pipe is not None:
            if self.idle_write_id:
                gobject.source_remove(self.idle_write_id)
                self.idle_write_id = 0

            if not self.tried_killing:
                os.kill(self.pipe.pid, signal.SIGTERM)
                self.tried_killing = True
            else:
                os.kill(self.pipe.pid, signal.SIGKILL)

    def on_child_end(self, pid, error_code):
        # In an idle, so it is emitted after all the std*-line signals
        # have been intercepted
        gobject.idle_add(self.emit, 'end-execute', error_code)

# ex:ts=4:et:



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys
import os
import shutil

import gtk
from gtk import gdk
import pluma
import platform

from WindowHelper import WindowHelper
from Library import Library
from Manager import Manager
from Snippet import Snippet

class SnippetsPlugin(pluma.Plugin):
        def __init__(self):
                pluma.Plugin.__init__(self)

                self.dlg = None

                library = Library()
                library.set_accelerator_callback(self.accelerator_activated)

                if platform.platform() == 'Windows':
                        snippetsdir = os.path.expanduser('~/pluma/snippets')
                else:
                        userdir = os.getenv('MATE22_USER_DIR')
                        if userdir:
                                snippetsdir = os.path.join(userdir, 'pluma/snippets')
                        else:
                                snippetsdir = os.path.expanduser('~/.config/pluma/snippets')

                library.set_dirs(snippetsdir, self.system_dirs())

        def system_dirs(self):
        	if platform.platform() != 'Windows':
		        if 'XDG_DATA_DIRS' in os.environ:
		                datadirs = os.environ['XDG_DATA_DIRS']
		        else:
		                datadirs = '/usr/local/share' + os.pathsep + '/usr/share'

		        dirs = []

		        for d in datadirs.split(os.pathsep):
		                d = os.path.join(d, 'pluma', 'plugins', 'snippets')

		                if os.path.isdir(d):
		                        dirs.append(d)

                dirs.append(self.get_data_dir())
                return dirs

        def activate(self, window):
                data = WindowHelper(self)
                window._snippets_plugin_data = data
                data.run(window)

        def deactivate(self, window):
                window._snippets_plugin_data.stop()
                window._snippets_plugin_data = None

        def update_ui(self, window):
                window._snippets_plugin_data.update()

        def create_configure_dialog(self):
                if not self.dlg:
                        self.dlg = Manager(self.get_data_dir())
                else:
                        self.dlg.run()

                window = pluma.app_get_default().get_active_window()

                if window:
                        self.dlg.dlg.set_transient_for(window)

                return self.dlg.dlg

        def accelerator_activated(self, group, obj, keyval, mod):
                ret = False

                if hasattr(obj, '_snippets_plugin_data'):
                        ret = obj._snippets_plugin_data.accelerator_activated(keyval, mod)

                return ret



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import tempfile
import shutil

import gobject
import gtk
from gtk import gdk
import gtksourceview2 as gsv
import pango
import pluma
import gio

from Snippet import Snippet
from Helper import *
from Library import *
from Importer import *
from Exporter import *
from Document import Document
from LanguageManager import get_language_manager

class Manager:
        NAME_COLUMN = 0
        SORT_COLUMN = 1
        OBJ_COLUMN = 2
        TARGET_URI = 105

        model = None
        drag_icons = ('mate-mime-application-x-tarz', 'mate-package', 'package')
        default_export_name = _('Snippets archive') + '.tar.gz'
        dragging = False
        dnd_target_list = [('text/uri-list', 0, TARGET_URI)]

        def __init__(self, datadir):
                self.datadir = datadir
                self.snippet = None
                self.dlg = None
                self._temp_export = None
                self.snippets_doc = None
                self.manager = None
                self.default_size = None

                self.key_press_id = 0
                self.run()
        
        def get_language_snippets(self, path, name = None):
                library = Library()
                
                name = self.get_language(path)
                nodes = library.get_snippets(name)

                return nodes

        def add_new_snippet_node(self, parent):
                return self.model.append(parent, ('<i>' + _('Add a new snippet...') + \
                                '</i>', '', None))

        def fill_language(self, piter, expand=True):
                # Remove all children
                child = self.model.iter_children(piter)
                
                while child and self.model.remove(child):
                        True
                
                path = self.model.get_path(piter)
                nodes = self.get_language_snippets(path)
                language = self.get_language(path)
                
                Library().ref(language)
                
                if nodes:
                        for node in nodes:
                                self.add_snippet(piter, node)
                else:
                        # Add node that tells there are no snippets currently
                        self.add_new_snippet_node(piter)

                if expand:
                        self.tree_view.expand_row(path, False)

        def build_model(self, force_reload = False):
                window = pluma.app_get_default().get_active_window()
                
                if window:
                        view = window.get_active_view()

                        if not view:
                                current_lang = None
                        else:
                                current_lang = view.get_buffer().get_language()
                                source_view = self['source_view_snippet']

                else:
                        current_lang = None

                tree_view = self['tree_view_snippets']
                expand = None
                
                if not self.model or force_reload:
                        self.model = gtk.TreeStore(str, str, object)
                        self.model.set_sort_column_id(self.SORT_COLUMN, gtk.SORT_ASCENDING)
                        manager = get_language_manager()
                        langs = pluma.language_manager_list_languages_sorted(manager, True)
                        
                        piter = self.model.append(None, (_('Global'), '', None))
                        # Add dummy node
                        self.model.append(piter, ('', '', None))
                        
                        nm = None
                        
                        if current_lang:
                                nm = current_lang.get_name()
                
                        for lang in langs:
                                name = lang.get_name()
                                parent = self.model.append(None, (name, name, lang))

                                # Add dummy node
                                self.model.append(parent, ('', '', None))

                                if (nm == name):
                                        expand = parent
                else:
                        if current_lang:
                                piter = self.model.get_iter_first()
                                nm = current_lang.get_name()
                                
                                while piter:
                                        lang = self.model.get_value(piter, \
                                                        self.SORT_COLUMN)
                                        
                                        if lang == nm:
                                                expand = piter
                                                break;
                                                
                                        piter = self.model.iter_next(piter)

                tree_view.set_model(self.model)
                
                if not expand:
                        expand = self.model.get_iter_root()
                        
                tree_view.expand_row(self.model.get_path(expand), False)
                self.select_iter(expand)

        def get_cell_data_pixbuf_cb(self, column, cell, model, iter):
                s = model.get_value(iter, self.OBJ_COLUMN)
                
                snippet = isinstance(s, SnippetData)
                
                if snippet and not s.valid:
                        cell.set_property('stock-id', gtk.STOCK_DIALOG_ERROR)
                else:
                        cell.set_property('stock-id', None)

                cell.set_property('xalign', 1.0)
                
        def get_cell_data_cb(self, column, cell, model, iter):
                s = model.get_value(iter, self.OBJ_COLUMN)
                
                snippet = isinstance(s, SnippetData)
                
                cell.set_property('editable', snippet)
                cell.set_property('markup', model.get_value(iter, self.NAME_COLUMN))

        def on_tree_view_drag_data_get(self, widget, context, selection_data, info, time):
                gfile = gio.File(self._temp_export)
                selection_data.set_uris([gfile.get_uri()])
       
        def on_tree_view_drag_begin(self, widget, context):
                self.dragging = True
                
                if self._temp_export:
                      shutil.rmtree(os.path.dirname(self._temp_export))
                      self._temp_export = None

                if self.dnd_name:
                        context.set_icon_name(self.dnd_name, 0, 0)

                dirname = tempfile.mkdtemp()
                filename = os.path.join(dirname, self.default_export_name)
                
                # Generate temporary file name
                self.export_snippets(filename, False)
                self._temp_export = filename
        
        def on_tree_view_drag_end(self, widget, context):
                self.dragging = False

        def on_tree_view_drag_data_received(self, widget, context, x, y, selection, info, timestamp):
                uris = selection.get_uris()
                
                self.import_snippets(uris)

        def on_tree_view_drag_motion(self, widget, context, x, y, timestamp):
                # Return False if we are dragging
                if self.dragging:
                        return False
                
                # Check uri target
                if not gtk.targets_include_uri(context.targets):
                        return False

                # Check action
                action = None
                if context.suggested_action == gdk.ACTION_COPY:
                        action = gdk.ACTION_COPY
                else:
                        for act in context.actions:
                                if act == gdk.ACTION_COPY:
                                      action = gdk.ACTION_COPY
                                      break  
                
                if action == gdk.ACTION_COPY:
                        context.drag_status(gdk.ACTION_COPY, timestamp)        
                        return True
                else:
                        return False

        def build_dnd(self):
                tv = self.tree_view
                
                # Set it as a drag source for exporting snippets
                tv.drag_source_set(gdk.BUTTON1_MASK, self.dnd_target_list, gdk.ACTION_DEFAULT | gdk.ACTION_COPY)
                
                # Set it as a drag destination for importing snippets
                tv.drag_dest_set(gtk.DEST_DEFAULT_HIGHLIGHT | gtk.DEST_DEFAULT_DROP, 
                                 self.dnd_target_list, gdk.ACTION_DEFAULT | gdk.ACTION_COPY)
                
                tv.connect('drag_data_get', self.on_tree_view_drag_data_get)
                tv.connect('drag_begin', self.on_tree_view_drag_begin)
                tv.connect('drag_end', self.on_tree_view_drag_end)
                tv.connect('drag_data_received', self.on_tree_view_drag_data_received)
                tv.connect('drag_motion', self.on_tree_view_drag_motion)

                theme = gtk.icon_theme_get_for_screen(tv.get_screen())
                
                self.dnd_name = None
                for name in self.drag_icons:
                        icon = theme.lookup_icon(name, gtk.ICON_SIZE_DND, 0)
                        
                        if icon:
                                self.dnd_name = name
                                break
                
        def build_tree_view(self):                
                self.tree_view = self['tree_view_snippets']
                
                self.column = gtk.TreeViewColumn(None)

                self.renderer = gtk.CellRendererText()
                self.column.pack_start(self.renderer, False)
                self.column.set_cell_data_func(self.renderer, self.get_cell_data_cb)

                renderer = gtk.CellRendererPixbuf()
                self.column.pack_start(renderer, True)
                self.column.set_cell_data_func(renderer, self.get_cell_data_pixbuf_cb)

                self.tree_view.append_column(self.column)
                
                self.renderer.connect('edited', self.on_cell_edited)
                self.renderer.connect('editing-started', self.on_cell_editing_started)

                selection = self.tree_view.get_selection()
                selection.set_mode(gtk.SELECTION_MULTIPLE)
                selection.connect('changed', self.on_tree_view_selection_changed)
                
                self.build_dnd()
        
        def build(self):
                self.builder = gtk.Builder()
                self.builder.add_from_file(os.path.join(self.datadir, 'ui', 'snippets.ui'))
                
                handlers_dic = {
                        'on_dialog_snippets_response': self.on_dialog_snippets_response,
                        'on_dialog_snippets_destroy': self.on_dialog_snippets_destroy,
                        'on_button_new_snippet_clicked': self.on_button_new_snippet_clicked,
                        'on_button_import_snippets_clicked': self.on_button_import_snippets_clicked,
                        'on_button_export_snippets_clicked': self.on_button_export_snippets_clicked,
                        'on_button_remove_snippet_clicked': self.on_button_remove_snippet_clicked,
                        'on_entry_tab_trigger_focus_out': self.on_entry_tab_trigger_focus_out,
                        'on_entry_tab_trigger_changed': self.on_entry_tab_trigger_changed,
                        'on_entry_accelerator_focus_out': self.on_entry_accelerator_focus_out,
                        'on_entry_accelerator_focus_in': self.on_entry_accelerator_focus_in,
                        'on_entry_accelerator_key_press': self.on_entry_accelerator_key_press,
                        'on_source_view_snippet_focus_out': self.on_source_view_snippet_focus_out,
                        'on_tree_view_snippets_row_expanded': self.on_tree_view_snippets_row_expanded,
                        'on_tree_view_snippets_key_press': self.on_tree_view_snippets_key_press}

                self.builder.connect_signals(handlers_dic)
                
                self.build_tree_view()
                self.build_model()

                image = self['image_remove']
                image.set_from_stock(gtk.STOCK_REMOVE, gtk.ICON_SIZE_SMALL_TOOLBAR)

                source_view = self['source_view_snippet']
                manager = get_language_manager()
                lang = manager.get_language('snippets')

                if lang:
                        source_view.get_buffer().set_highlight_syntax(True)
                        source_view.get_buffer().set_language(lang)
                        self.snippets_doc = Document(None, source_view)

                combo = self['combo_drop_targets']
                combo.set_text_column(0)

                entry = combo.child
                entry.connect('focus-out-event', self.on_entry_drop_targets_focus_out)
                entry.connect('drag-data-received', self.on_entry_drop_targets_drag_data_received)
                
                lst = entry.drag_dest_get_target_list()
                lst = gtk.target_list_add_uri_targets(entry.drag_dest_get_target_list(), self.TARGET_URI)
                entry.drag_dest_set_target_list(lst)
                
                self.dlg = self['dialog_snippets']
                
                if self.default_size:
                        self.dlg.set_default_size(*self.default_size)
        
        def __getitem__(self, key):
                return self.builder.get_object(key)

        def is_filled(self, piter):
                if not self.model.iter_has_child(piter):
                        return True
                
                child = self.model.iter_children(piter)
                nm = self.model.get_value(child, self.NAME_COLUMN)
                obj = self.model.get_value(child, self.OBJ_COLUMN)
                
                return (obj or nm)

        def fill_if_needed(self, piter, expand=True):
                if not self.is_filled(piter):
                        self.fill_language(piter, expand)

        def find_iter(self, parent, snippet):
                self.fill_if_needed(parent)
                piter = self.model.iter_children(parent)
                
                while (piter):
                        node = self.model.get_value(piter, self.OBJ_COLUMN)

                        if node == snippet.data:
                                return piter
                        
                        piter = self.model.iter_next(piter)
                
                return None

        def selected_snippets_state(self):
                snippets = self.selected_snippets(False)
                override = False
                remove = False
                system = False
                
                for snippet in snippets:
                        if not snippet:
                                continue

                        if snippet.is_override():
                                override = True
                        elif snippet.can_modify():
                                remove = True
                        else:
                                system = True
                        
                        # No need to continue if both are found
                        if override and remove:
                                break

                return (override, remove, system)

        def update_buttons(self):
                button_remove = self['button_remove_snippet']
                button_new = self['button_new_snippet']
                image_remove = self['image_remove']

                button_new.set_sensitive(self.language_path != None)
                override, remove, system = self.selected_snippets_state()
                
                if not (override ^ remove) or system:
                        button_remove.set_sensitive(False)
                        image_remove.set_from_stock(gtk.STOCK_DELETE, gtk.ICON_SIZE_BUTTON)
                else:
                        button_remove.set_sensitive(True)
                        
                        if override:
                                image_remove.set_from_stock(gtk.STOCK_UNDO, gtk.ICON_SIZE_BUTTON)
                                tooltip = _('Revert selected snippet')
                        else:
                                image_remove.set_from_stock(gtk.STOCK_DELETE, gtk.ICON_SIZE_BUTTON)
                                tooltip = _('Delete selected snippet')
                        
                        button_remove.set_tooltip_text(tooltip)

        def snippet_changed(self, piter = None):
                if piter:
                        node = self.model.get_value(piter, self.OBJ_COLUMN)
                        s = Snippet(node)
                else:
                        s = self.snippet
                        piter = self.find_iter(self.model.get_iter(self.language_path), s)

                if piter:
                        nm = s.display()
                        
                        self.model.set(piter, self.NAME_COLUMN, nm, self.SORT_COLUMN, nm)
                        self.update_buttons()
                        self.entry_tab_trigger_update_valid()

                return piter

        def add_snippet(self, parent, snippet):
                piter = self.model.append(parent, ('', '', snippet))
                
                return self.snippet_changed(piter)

        def run(self):
                if not self.dlg:
                        self.build()
                        self.dlg.show()
                else:
                        self.build_model()
                        self.dlg.present()

        
        def snippet_from_iter(self, model, piter):
                parent = model.iter_parent(piter)
                
                if parent:
                        return model.get_value(piter, self.OBJ_COLUMN)
                else:
                        return None
        
        def language_snippets(self, model, parent, as_path=False):
                self.fill_if_needed(parent, False)
                piter = model.iter_children(parent)
                snippets = []
                
                if not piter:
                        return snippets
                
                while piter:
                        snippet = self.snippet_from_iter(model, piter)
                        
                        if snippet:
                                if as_path:
                                        snippets.append(model.get_path(piter))
                                else:
                                        snippets.append(snippet)

                        piter = model.iter_next(piter)
                
                return snippets
        
        def selected_snippets(self, include_languages=True, as_path=False):
                selection = self.tree_view.get_selection()
                (model, paths) = selection.get_selected_rows()
                snippets = []
                
                if paths and len(paths) != 0:
                        for p in paths:
                                piter = model.get_iter(p)
                                parent = model.iter_parent(piter)
                                
                                if not piter:
                                        continue
                                
                                if parent:
                                        snippet = self.snippet_from_iter(model, piter)
                                        
                                        if not snippet:
                                                continue
                                        
                                        if as_path:
                                                snippets.append(p)
                                        else:
                                                snippets.append(snippet)
                                elif include_languages:
                                        snippets += self.language_snippets(model, piter, as_path)
                        
                return snippets                        
        
        def selected_snippet(self):
                selection = self.tree_view.get_selection()
                (model, paths) = selection.get_selected_rows()
                
                if len(paths) == 1:
                        piter = model.get_iter(paths[0])
                        parent = model.iter_parent(piter)
                        snippet = self.snippet_from_iter(model, piter)
                        
                        return parent, piter, snippet
                else:
                        return None, None, None

        def selection_changed(self):
                if not self.snippet:
                        sens = False

                        self['entry_tab_trigger'].set_text('')
                        self['entry_accelerator'].set_text('')
                        buf = self['source_view_snippet'].get_buffer()
                        buf.begin_not_undoable_action()
                        buf.set_text('')
                        buf.end_not_undoable_action()
                        self['combo_drop_targets'].child.set_text('')

                else:
                        sens = True

                        self['entry_tab_trigger'].set_text(self.snippet['tag'])
                        self['entry_accelerator'].set_text( \
                                        self.snippet.accelerator_display())
                        self['combo_drop_targets'].child.set_text(', '.join(self.snippet['drop-targets']))
                        
                        buf = self['source_view_snippet'].get_buffer()
                        buf.begin_not_undoable_action()
                        buf.set_text(self.snippet['text'])
                        buf.end_not_undoable_action()


                for name in ['source_view_snippet', 'label_tab_trigger',
                                'entry_tab_trigger', 'label_accelerator', 
                                'entry_accelerator', 'label_drop_targets',
                                'combo_drop_targets']:
                        self[name].set_sensitive(sens)
                
                self.update_buttons()
                        
        def select_iter(self, piter, unselect=True):
                selection = self.tree_view.get_selection()
                
                if unselect:
                        selection.unselect_all()

                selection.select_iter(piter)
                
                self.tree_view.scroll_to_cell(self.model.get_path(piter), None, \
                        True, 0.5, 0.5)

        def get_language(self, path):
                if path[0] == 0:
                        return None
                else:
                        return self.model.get_value(self.model.get_iter( \
                                        (path[0],)), self.OBJ_COLUMN).get_id()

        def new_snippet(self, properties=None):
                if not self.language_path:
                        return None

                snippet = Library().new_snippet(self.get_language(self.language_path), properties)
                
                return Snippet(snippet)

        def get_dummy(self, parent):
                if not self.model.iter_n_children(parent) == 1:
                        return None
                
                dummy = self.model.iter_children(parent)
                
                if not self.model.get_value(dummy, self.OBJ_COLUMN):
                        return dummy
        
                return None
        
        def unref_languages(self):
                piter = self.model.get_iter_first()
                library = Library()
                
                while piter:
                        if self.is_filled(piter):
                                language = self.get_language(self.model.get_path(piter))
                                library.save(language)

                                library.unref(language)
                        
                        piter = self.model.iter_next(piter)

        # Callbacks
        def on_dialog_snippets_destroy(self, dlg):
                # Remove temporary drag export
                if self._temp_export:
                      shutil.rmtree(os.path.dirname(self._temp_export))
                      self._temp_export = None

                if self.snippets_doc:
                        self.snippets_doc.stop()
                
                self.default_size = [dlg.allocation.width, dlg.allocation.height]
                self.manager = None

                self.unref_languages()        
                self.snippet = None        
                self.model = None
                self.dlg = None                
        
        def on_dialog_snippets_response(self, dlg, resp):                                
                if resp == gtk.RESPONSE_HELP:
                        pluma.help_display(self.dlg, 'pluma', 'pluma-snippets-plugin')
                        return

                self.dlg.destroy()
        
        def on_cell_editing_started(self, renderer, editable, path):
                piter = self.model.get_iter(path)
                
                if not self.model.iter_parent(piter):
                        renderer.stop_editing(True)
                        editable.remove_widget()
                elif isinstance(editable, gtk.Entry):
                        if self.snippet:
                                editable.set_text(self.snippet['description'])
                        else:
                                # This is the `Add a new snippet...` item
                                editable.set_text('')
                        
                        editable.grab_focus()
        
        def on_cell_edited(self, cell, path, new_text):                
                if new_text != '':
                        piter = self.model.get_iter(path)
                        node = self.model.get_value(piter, self.OBJ_COLUMN)
                        
                        if node:
                                if node == self.snippet.data:
                                        s = self.snippet
                                else:
                                        s = Snippet(node)
                        
                                s['description'] = new_text
                                self.snippet_changed(piter)
                                self.select_iter(piter)
                        else:
                                # This is the `Add a new snippet...` item
                                # We create a new snippet
                                snippet = self.new_snippet({'description': new_text})
                                
                                if snippet:
                                        self.model.set(piter, self.OBJ_COLUMN, snippet.data)
                                        self.snippet_changed(piter)
                                        self.snippet = snippet
                                        self.selection_changed()
        
        def on_entry_accelerator_focus_out(self, entry, event):
                if not self.snippet:
                        return

                entry.set_text(self.snippet.accelerator_display())

        def entry_tab_trigger_update_valid(self):
                entry = self['entry_tab_trigger']
                text = entry.get_text()
                
                if text and not Library().valid_tab_trigger(text):
                        img = self['image_tab_trigger']
                        img.set_from_stock(gtk.STOCK_DIALOG_ERROR, gtk.ICON_SIZE_BUTTON)
                        img.show()

                        #self['hbox_tab_trigger'].set_spacing(3)
                        tip = _('This is not a valid Tab trigger. Triggers can either contain letters or a single (non-alphanumeric) character like: {, [, etc.')
                        
                        entry.set_tooltip_text(tip)
                        img.set_tooltip_text(tip)
                else:
                        self['image_tab_trigger'].hide()
                        #self['hbox_tab_trigger'].set_spacing(0)
                        entry.set_tooltip_text(_('Single word the snippet is activated with after pressing Tab'))
                
                return False

        def on_entry_tab_trigger_focus_out(self, entry, event):
                if not self.snippet:
                        return

                text = entry.get_text()

                # save tag
                self.snippet['tag'] = text
                self.snippet_changed()
        
        def on_entry_drop_targets_focus_out(self, entry, event):
                if not self.snippet:
                        return
                
                text = entry.get_text()

                # save drop targets
                self.snippet['drop-targets'] = text
                self.snippet_changed()
        
        def on_entry_tab_trigger_changed(self, entry):
                self.entry_tab_trigger_update_valid()
        
        def on_source_view_snippet_focus_out(self, source_view, event):
                if not self.snippet:
                        return

                buf = source_view.get_buffer()
                text = buf.get_text(buf.get_start_iter(), \
                                buf.get_end_iter())

                self.snippet['text'] = text
                self.snippet_changed()
        
        def on_button_new_snippet_clicked(self, button):
                snippet = self.new_snippet()
                
                if not snippet:
                        return

                parent = self.model.get_iter(self.language_path)
                path = self.model.get_path(parent)
                
                dummy = self.get_dummy(parent)
                
                if dummy:
                        # Remove the dummy
                        self.model.remove(dummy)
                
                # Add the snippet
                piter = self.add_snippet(parent, snippet.data)
                self.select_iter(piter)

                if not self.tree_view.row_expanded(path):
                        self.tree_view.expand_row(path, False)
                        self.select_iter(piter)

                self.tree_view.grab_focus()

                path = self.model.get_path(piter)
                self.tree_view.set_cursor(path, self.column, True)
        
        def file_filter(self, name, pattern):
                fil = gtk.FileFilter()
                fil.set_name(name)
                
                for p in pattern:
                        fil.add_pattern(p)
                
                return fil
        
        def import_snippets(self, filenames):
                success = True
                
                for filename in filenames:
                        if not pluma.utils.uri_has_file_scheme(filename):
                                continue

                        # Remove file://
                        gfile = gio.File(filename)
                        filename = gfile.get_path()

                        importer = Importer(filename)
                        error = importer.run()
         
                        if error:
                                message = _('The following error occurred while importing: %s') % error
                                success = False
                                message_dialog(self.dlg, gtk.MESSAGE_ERROR, message)
                
                self.build_model(True)

                if success:
                        message = _('Import successfully completed')
                        message_dialog(self.dlg, gtk.MESSAGE_INFO, message)
               
        def on_import_response(self, dialog, response):
                if response == gtk.RESPONSE_CANCEL or response == gtk.RESPONSE_CLOSE:
                        dialog.destroy()
                        return
                
                f = dialog.get_uris()
                dialog.destroy()
                
                self.import_snippets(f)
                
        def on_button_import_snippets_clicked(self, button):
                dlg = gtk.FileChooserDialog(parent=self.dlg, title=_("Import snippets"), 
                                action=gtk.FILE_CHOOSER_ACTION_OPEN, 
                                buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, 
                                         gtk.STOCK_OPEN, gtk.RESPONSE_OK))
                
                dlg.add_filter(self.file_filter(_('All supported archives'), ('*.gz','*.bz2','*.tar', '*.xml')))
                dlg.add_filter(self.file_filter(_('Gzip compressed archive'), ('*.tar.gz',)))
                dlg.add_filter(self.file_filter(_('Bzip2 compressed archive'), ('*.tar.bz2',)))
                dlg.add_filter(self.file_filter(_('Single snippets file'), ('*.xml',)))
                dlg.add_filter(self.file_filter(_('All files'), '*'))

                dlg.connect('response', self.on_import_response)
                dlg.set_local_only(True)
                
                dlg.show()

        def export_snippets_real(self, filename, snippets, show_dialogs=True):
                export = Exporter(filename, snippets)
                error = export.run()
                
                if error:
                        message = _('The following error occurred while exporting: %s') % error
                        msgtype = gtk.MESSAGE_ERROR
                        retval = False
                else:
                        message = _('Export successfully completed')
                        msgtype = gtk.MESSAGE_INFO
                        retval = True

                if show_dialogs:
                        message_dialog(self.dlg, msgtype, message)

                return retval
                
        def on_export_response(self, dialog, response):
                filename = dialog.get_filename()
                snippets = dialog._export_snippets
                
                dialog.destroy()
                
                if response != gtk.RESPONSE_OK:
                        return
                
                self.export_snippets_real(filename, snippets);
        
        def export_snippets(self, filename=None, show_dialogs=True):
                snippets = self.selected_snippets()
                
                if not snippets or len(snippets) == 0:
                        return False
                        
                usersnippets = []
                systemsnippets = []

                # Iterate through snippets and look for system snippets
                for snippet in snippets:
                        if snippet.can_modify():
                                usersnippets.append(snippet)
                        else:
                                systemsnippets.append(snippet)
               
                export_snippets = snippets

                if len(systemsnippets) != 0 and show_dialogs:
                        # Ask if system snippets should also be exported
                        message = _('Do you want to include selected <b>system</b> snippets in your export?')
                        mes = gtk.MessageDialog(flags=gtk.DIALOG_MODAL, 
                                        type=gtk.MESSAGE_QUESTION, 
                                        buttons=gtk.BUTTONS_YES_NO,
                                        message_format=message)
                        mes.set_property('use-markup', True)
                        resp = mes.run()
                        mes.destroy()
                        
                        if resp == gtk.RESPONSE_NO:
                                export_snippets = usersnippets
                        elif resp != gtk.RESPONSE_YES:
                                return False
                
                if len(export_snippets) == 0 and show_dialogs:                        
                        message = _('There are no snippets selected to be exported')
                        message_dialog(self.dlg, gtk.MESSAGE_INFORMATION, message)
                        return False
                
                if not filename:
                        dlg = gtk.FileChooserDialog(parent=self.dlg, title=_('Export snippets'), 
                                        action=gtk.FILE_CHOOSER_ACTION_SAVE, 
                                        buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, 
                                                 gtk.STOCK_SAVE, gtk.RESPONSE_OK))
                        
                        dlg._export_snippets = export_snippets
                        dlg.add_filter(self.file_filter(_('All supported archives'), ('*.gz','*.bz2','*.tar')))
                        dlg.add_filter(self.file_filter(_('Gzip compressed archive'), ('*.tar.gz',)))
                        dlg.add_filter(self.file_filter(_('Bzip2 compressed archive'), ('*.tar.bz2',)))

                        dlg.add_filter(self.file_filter(_('All files'), '*'))
                        dlg.set_do_overwrite_confirmation(True)
                        dlg.set_current_name(self.default_export_name)
                
                        dlg.connect('response', self.on_export_response)
                        dlg.set_local_only(True)
                
                        dlg.show()
                        return True
                else:
                        return self.export_snippets_real(filename, export_snippets, show_dialogs)
        
        def on_button_export_snippets_clicked(self, button):
                snippets = self.selected_snippets()
                
                if not snippets or len(snippets) == 0:
                        return
                        
                usersnippets = []
                systemsnippets = []

                # Iterate through snippets and look for system snippets
                for snippet in snippets:
                        if snippet.can_modify():
                                usersnippets.append(snippet)
                        else:
                                systemsnippets.append(snippet)

                dlg = gtk.FileChooserDialog(parent=self.dlg, title=_('Export snippets'), 
                                action=gtk.FILE_CHOOSER_ACTION_SAVE, 
                                buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, 
                                         gtk.STOCK_SAVE, gtk.RESPONSE_OK))
                
                dlg._export_snippets = snippets

                if len(systemsnippets) != 0:
                        # Ask if system snippets should also be exported
                        message = _('Do you want to include selected <b>system</b> snippets in your export?')
                        mes = gtk.MessageDialog(flags=gtk.DIALOG_MODAL, 
                                        type=gtk.MESSAGE_QUESTION, 
                                        buttons=gtk.BUTTONS_YES_NO,
                                        message_format=message)
                        mes.set_property('use-markup', True)
                        resp = mes.run()
                        mes.destroy()
                        
                        if resp == gtk.RESPONSE_NO:
                                dlg._export_snippets = usersnippets
                        elif resp != gtk.RESPONSE_YES:
                                dlg.destroy()
                                return
                
                if len(dlg._export_snippets) == 0:
                        dlg.destroy()
                        
                        message = _('There are no snippets selected to be exported')
                        message_dialog(self.dlg, gtk.MESSAGE_INFORMATION, message)
                        return
                
                dlg.add_filter(self.file_filter(_('All supported archives'), ('*.gz','*.bz2','*.tar')))
                dlg.add_filter(self.file_filter(_('Gzip compressed archive'), ('*.tar.gz',)))
                dlg.add_filter(self.file_filter(_('Bzip2 compressed archive'), ('*.tar.bz2',)))

                dlg.add_filter(self.file_filter(_('All files'), '*'))
                dlg.set_do_overwrite_confirmation(True)
                dlg.set_current_name(self.default_export_name)
                
                dlg.connect('response', self.on_export_response)
                dlg.set_local_only(True)
                
                dlg.show()                
        
        def remove_snippet_revert(self, path, piter):
                node = self.snippet_from_iter(self.model, piter)
                Library().revert_snippet(node)
                
                return piter
        
        def remove_snippet_delete(self, path, piter):
                node = self.snippet_from_iter(self.model, piter)
                parent = self.model.iter_parent(piter)

                Library().remove_snippet(node)

                if self.model.remove(piter):
                        return piter
                elif path[-1] != 0:
                        self.select_iter(self.model.get_iter((path[0], path[1] - 1)))
                else:
                        dummy = self.add_new_snippet_node(parent)
                        self.tree_view.expand_row(self.model.get_path(parent), False)
                        return dummy
       
        def on_button_remove_snippet_clicked(self, button):
                override, remove, system = self.selected_snippets_state()
                
                if not (override ^ remove) or system:
                        return
                
                paths = self.selected_snippets(include_languages=False, as_path=True)
                
                if override:
                        action = self.remove_snippet_revert
                else:
                        action = self.remove_snippet_delete
                
                # Remove selection
                self.tree_view.get_selection().unselect_all()
                
                # Create tree row references
                references = []
                for path in paths:
                        references.append(gtk.TreeRowReference(self.model, path))

                # Remove/revert snippets
                select = None
                for reference in references:
                        path = reference.get_path()
                        piter = self.model.get_iter(path)
                        
                        res = action(path, piter)
                        
                        if res:
                                select = res

                if select:
                        self.select_iter(select)

                self.selection_changed()
        
        def set_accelerator(self, keyval, mod):
                accelerator = gtk.accelerator_name(keyval, mod)
                self.snippet['accelerator'] = accelerator

                return True
        
        def on_entry_accelerator_key_press(self, entry, event):
                source_view = self['source_view_snippet']

                if event.keyval == gdk.keyval_from_name('Escape'):
                        # Reset
                        entry.set_text(self.snippet.accelerator_display())
                        self.tree_view.grab_focus()
                        
                        return True
                elif event.keyval == gdk.keyval_from_name('Delete') or \
                                event.keyval == gdk.keyval_from_name('BackSpace'):
                        # Remove the accelerator
                        entry.set_text('')
                        self.snippet['accelerator'] = ''
                        self.tree_view.grab_focus()
                        
                        self.snippet_changed()
                        return True
                elif Library().valid_accelerator(event.keyval, event.state):
                        # New accelerator
                        self.set_accelerator(event.keyval, \
                                        event.state & gtk.accelerator_get_default_mod_mask())
                        entry.set_text(self.snippet.accelerator_display())
                        self.snippet_changed()
                        self.tree_view.grab_focus()

                else:
                        return True
        
        def on_entry_accelerator_focus_in(self, entry, event):
                if self.snippet['accelerator']:
                        entry.set_text(_('Type a new shortcut, or press Backspace to clear'))
                else:
                        entry.set_text(_('Type a new shortcut'))
        
        def update_language_path(self):
                model, paths = self.tree_view.get_selection().get_selected_rows()
                
                # Check if all have the same language parent
                current_parent = None

                for path in paths:
                        piter = model.get_iter(path)
                        parent = model.iter_parent(piter)
                        
                        if parent:
                                path = model.get_path(parent)

                        if current_parent != None and current_parent != path:
                                current_parent = None
                                break
                        else:
                                current_parent = path

                self.language_path = current_parent
                
        def on_tree_view_selection_changed(self, selection):
                parent, piter, node = self.selected_snippet()
                
                if self.snippet:
                        self.on_entry_tab_trigger_focus_out(self['entry_tab_trigger'],
                                        None)
                        self.on_source_view_snippet_focus_out(self['source_view_snippet'], 
                                        None)
                        self.on_entry_drop_targets_focus_out(self['combo_drop_targets'].child,
                                        None)
                
                self.update_language_path()

                if node:
                        self.snippet = Snippet(node)
                else:
                        self.snippet = None

                self.selection_changed()

        def iter_after(self, target, after):
                if not after:
                        return True

                tp = self.model.get_path(target)
                ap = self.model.get_path(after)
                
                if tp[0] > ap[0] or (tp[0] == ap[0] and (len(ap) == 1 or tp[1] > ap[1])):
                        return True
                
                return False
                
        def on_tree_view_snippets_key_press(self, treeview, event):
                if event.keyval == gdk.keyval_from_name('Delete'):
                        self.on_button_remove_snippet_clicked(None)
                        return True

        def on_tree_view_snippets_row_expanded(self, treeview, piter, path):
                # Check if it is already filled
                self.fill_if_needed(piter)
                self.select_iter(piter)
        
        def on_entry_drop_targets_drag_data_received(self, entry, context, x, y, selection_data, info, timestamp):
                if not gtk.targets_include_uri(context.targets):
                        return
                
                uris = drop_get_uris(selection_data)
                
                if not uris:
                        return
                
                if entry.get_text():
                        mimes = [entry.get_text()]
                else:
                        mimes = []
                
                for uri in uris:
                        try:
                                mime = gio.content_type_guess(uri)
                        except:
                                mime = None
                        
                        if mime:
                                mimes.append(mime)
                
                entry.set_text(', '.join(mimes))
                self.on_entry_drop_targets_focus_out(entry, None)
                context.finish(True, False, timestamp)
                
                entry.stop_emission('drag_data_received')
# ex:ts=8:et:



# -*- coding: utf-8 -*-

#  Copyright (C) 2009 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import pluma
from windowhelper import WindowHelper

class QuickOpenPlugin(pluma.Plugin):
        def __init__(self):
                pluma.Plugin.__init__(self)

                self._popup_size = (450, 300)
                self._helpers = {}

        def activate(self, window):
                self._helpers[window] = WindowHelper(window, self)

        def deactivate(self, window):
                self._helpers[window].deactivate()
                del self._helpers[window]

        def update_ui(self, window):
                self._helpers[window].update_ui()

        def get_popup_size(self):
                return self._popup_size

        def set_popup_size(self, size):
                self._popup_size = size

# ex:ts=8:et:



# -*- coding: utf-8 -*-

# preprocessor.py - simple preprocessor for plugin template files
# This file is part of pluma
#
# Copyright (C) 2006 - Steve Frécinaux
#
# pluma is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pluma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pluma; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, 
# Boston, MA  02110-1301  USA

import sys
import re

class DeepnessException(Exception):
    def __init__(self):
        Exception.__init__(self)

statements = [re.compile("^##\s*%s\s*$" % pattern) for pattern in
                        ['(?P<stmt>ifdef|ifndef)\s+(?P<key>[^\s]+)',
                         '(?P<stmt>elif|if)\s+(?P<expr>.+)',
                         '(?P<stmt>else|endif)',
                         '(?P<stmt>define)\s+(?P<key>[^\s]+)(\s+(?P<val>.+))?',
                         '(?P<stmt>undef)\s+(?P<key>[^\s]+)']]
variable = re.compile("##\((?P<name>[a-zA-Z_][a-zA-Z0-9_]*)(?P<mods>(\.[a-z]+)+)?\)")


def _eval(expr, macros):
    return eval(expr,
                {'defined': lambda x: macros.has_key(x)},
                macros)

def _subvar(match, macros):
    name = match.group('name')
    if name in macros:
        val = str(macros[name])
        if val is None:
            return ''
    else:
        return ''
    
    mods = match.group('mods')
    if mods is not None:
        for mod in mods[1:].split('.'):
            if mod == 'lower':
                val = val.lower()
            elif mod == 'upper':
                val = val.upper()
            elif mod == 'camel':
                val = ''.join(i.capitalize() 
                              for i in val.split('_'))
    return val

def process(infile = sys.stdin, outfile = sys.stdout, macros = {}):
    if not isinstance(infile, file):
        infile = open(infile, mode = 'r')
        close_infile = True
    else:
        close_infile = False

    if not isinstance(outfile, file):
        outfile = open(outfile, mode = 'w')
        close_outfile = True
    else:
        close_outfile = False
    
    deepness = 0
    writing_disabled = None
    
    for line in infile:
        # Skip comments
        if line[0:3].lower() == '##c':
            continue

        # Check whether current line is a preprocessor directive        
        for statement in statements:
            match = statement.match(line)
            if match: break
    
        if match is not None:
            stmt = match.group('stmt')

            if stmt == "define":
                if writing_disabled is None:
                    key = match.group('key')
                    val = match.group('val')
                    macros[key] = val

            elif stmt == "undef":
                if writing_disabled is None:
                    key = match.group('key')
                    if key in macros:
                        del macros[key]

            elif stmt == "ifdef":
                deepness += 1
                if writing_disabled is None and \
                   match.group('key') not in macros:
                    writing_disabled = deepness

            elif stmt == "ifndef":
                deepness += 1
                if writing_disabled is None and \
                   match.group('key') in macros:
                    writing_disabled = deepness                    

            elif stmt == "if":
                deepness += 1
                if writing_disabled is None and \
                   not _eval(match.group('expr'), macros):
                       writing_disabled = deepness

            elif stmt == "elif":
                if deepness == 0:
                    raise DeepnessException()
                if writing_disabled is None and \
                   not _eval(match.group('expr'), macros):
                    writing_disabled = deepness
                elif writing_disabled == deepness:
                    writing_disabled = None

            elif stmt == "else":
                if deepness == 0:
                    raise DeepnessException()
                if writing_disabled is None:
                    writing_disabled = deepness
                elif writing_disabled == deepness:
                    writing_disabled = None

            elif stmt == "endif":
                if deepness == 0:
                    raise DeepnessException()
                if writing_disabled is not None and \
                   writing_disabled == deepness:
                    writing_disabled = None
                deepness -= 1

        # Do variable substitution in the remaining lines
        elif writing_disabled is None:
            outfile.write(re.sub(variable, 
                                 lambda m: _subvar(m, macros),
                                 line))

    if deepness != 0:
        raise DeepnessException()

    if close_infile: infile.close()
    if close_outfile: outfile.close()

# ex:ts=4:et:



#    Pluma snippets plugin
#    Copyright (C) 2006-2007  Jesse van den Kieboom <jesse@icecrew.nl>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import re

class ParseError(Exception):
        def __str__(self):
                return 'Parse error, resume next'

class Modifiers:
        def _first_char(s):
                first = (s != '' and s[0]) or ''
                rest = (len(s) > 1 and s[1:]) or ''
                
                return first, rest
        
        def upper_first(s):
                first, rest = Modifiers._first_char(s)

                return '%s%s' % (first.upper(), rest)

        def upper(s):
                return s.upper()

        def lower_first(s):
                first, rest = Modifiers._first_char(s)

                return '%s%s' % (first.lower(), rest)
                
        def lower(s):
                return s.lower()
                
        def title(s):
                return s.title()
        
        upper_first = staticmethod(upper_first)
        upper = staticmethod(upper)
        lower_first = staticmethod(lower_first)
        lower = staticmethod(lower)
        title = staticmethod(title)
        _first_char = staticmethod(_first_char)

class SubstitutionParser:
        REG_ID = '[0-9]+'
        REG_NAME = '[a-zA-Z_]+'
        REG_MOD = '[a-zA-Z]+'
        REG_ESCAPE = '\\\\|\\(\\?|,|\\)'

        def __init__(self, pattern, groups = {}, modifiers = {}):
                self.pattern = pattern
                self.groups = groups
                
                self.REG_GROUP = '(?:(%s)|<(%s|%s)(?:,(%s))?>)' % (self.REG_ID, self.REG_ID, self.REG_NAME, self.REG_MOD)
                self.modifiers = {'u': Modifiers.upper_first,
                                  'U': Modifiers.upper,
                                  'l': Modifiers.lower_first,
                                  'L': Modifiers.lower,
                                  't': Modifiers.title}
                
                for k, v in modifiers.items():
                        self.modifiers[k] = v
        
        def parse(self):
                result, tokens = self._parse(self.pattern, None)
                
                return result
        
        def _parse(self, tokens, terminator):
                result = ''

                while tokens != '':
                        if self._peek(tokens) == '' or self._peek(tokens) == terminator:
                                tokens = self._remains(tokens)
                                break

                        try:
                                res, tokens = self._expr(tokens, terminator)
                        except ParseError:
                                res, tokens = self._text(tokens)
                        
                        result += res
                                
                return result, tokens
       
        def _peek(self, tokens, num = 0):
                return (num < len(tokens) and tokens[num])
        
        def _token(self, tokens):
                if tokens == '':
                        return '', '';

                return tokens[0], (len(tokens) > 1 and tokens[1:]) or ''
        
        def _remains(self, tokens, num = 1):
                return (num < len(tokens) and tokens[num:]) or ''

        def _expr(self, tokens, terminator):                
                if tokens == '':
                        return ''

                try:
                        return {'\\': self._escape,
                                '(': self._condition}[self._peek(tokens)](tokens, terminator)
                except KeyError:
                        raise ParseError

        def _text(self, tokens):
                return self._token(tokens)

        def _substitute(self, group, modifiers = ''):
                result = (self.groups.has_key(group) and self.groups[group]) or ''
                
                for modifier in modifiers:
                        if self.modifiers.has_key(modifier):
                                result = self.modifiers[modifier](result)
                
                return result
                
        def _match_group(self, tokens):
                match = re.match('\\\\%s' % self.REG_GROUP, tokens)
                
                if not match:
                        return None, tokens
                
                return self._substitute(match.group(1) or match.group(2), match.group(3) or ''), tokens[match.end():]
                
        def _escape(self, tokens, terminator):
                # Try to match a group
                result, tokens = self._match_group(tokens)
                
                if result != None:
                        return result, tokens
                
                s = self.REG_GROUP
                
                if terminator:
                        s += '|%s' % re.escape(terminator)

                match = re.match('\\\\(\\\\%s|%s)' % (s, self.REG_ESCAPE), tokens)

                if not match:
                        raise ParseError
                               
                return match.group(1), tokens[match.end():]  
        
        def _condition_value(self, tokens):
                match = re.match('\\\\?%s\s*' % self.REG_GROUP, tokens)
                
                if not match:
                        return None, tokens

                groups = match.groups()
                name = groups[0] or groups[1]

                return self.groups.has_key(name) and self.groups[name] != None, tokens[match.end():]        
        
        def _condition(self, tokens, terminator):
                # Match ? after (
                if self._peek(tokens, 1) != '?':
                        raise ParseError
               
                # Remove initial (? token
                tokens = self._remains(tokens, 2)
                condition, tokens = self._condition_value(tokens)
                
                if condition == None or self._peek(tokens) != ',':
                        raise ParseError

                truepart, tokens = self._parse(self._remains(tokens), ',')
                
                if truepart == None:
                        raise ParseError
                        
                falsepart, tokens = self._parse(tokens, ')')
                
                if falsepart == None:
                        raise ParseError

                if condition:
                        return truepart, tokens
                else:
                        return falsepart, tokens
        
        def escape_substitution(substitution):
                return re.sub('(%s|%s)' % (self.REG_GROUP, self.REG_ESCAPE), '\\\\\\1', substitution)
                
        escapesubstitution = staticmethod(escape_substitution)
# ex:ts=8:et:



import os
import tempfile
import sys
import shutil

from snippets.Library import *

class Importer:
        def __init__(self, filename):
                self.filename = filename

        def import_destination(self, filename):
                userdir = Library().userdir
                
                filename = os.path.basename(filename)
                (root, ext) = os.path.splitext(filename)

                filename = os.path.join(userdir, root + ext)
                i = 1
                
                while os.path.exists(filename):
                        filename = os.path.join(userdir, root + '_' + str(i) + ext)
                        i += 1
                
                return filename
                
        def import_file(self, filename):
                if not os.path.exists(filename):
                        return _('File "%s" does not exist') % filename
                
                if not os.path.isfile(filename):
                        return _('File "%s" is not a valid snippets file') % filename
                
                # Find destination for file to copy to
                dest = self.import_destination(filename)

                # Copy file
                shutil.copy(filename, dest)
                
                # Add library
                if not Library().add_user_library(dest):
                        return _('Imported file "%s" is not a valid snippets file') % os.path.basename(dest)
        
        def import_xml(self):
                return self.import_file(self.filename)

        def import_archive(self, cmd):
                dirname = tempfile.mkdtemp()
                status = os.system('cd %s; %s "%s"' % (dirname, cmd, self.filename))
                
                if status != 0:
                        return _('The archive "%s" could not be extracted' % self.filename)
                
                errors = []

                # Now import all the files from the archive
                for f in os.listdir(dirname):
                        f = os.path.join(dirname, f)

                        if os.path.isfile(f):
                                if self.import_file(f):
                                        errors.append(os.path.basename(f))
                        else:
                                sys.stderr.write('Skipping %s, not a valid snippets file' % os.path.basename(f))
                
                # Remove the temporary directory
                shutil.rmtree(dirname)

                if len(errors) > 0:
                        return _('The following files could not be imported: %s') % ', '.join(errors)
                        
        def import_targz(self):
                self.import_archive('tar -x --gzip -f')
        
        def import_tarbz2(self):
                self.import_archive('tar -x --bzip2 -f')

        def import_tar(self):
                self.import_archive('tar -xf')
        
        def run(self):
                if not os.path.exists(self.filename):
                        return _('File "%s" does not exist') % self.filename
                
                if not os.path.isfile(self.filename):
                        return _('File "%s" is not a valid snippets archive') % self.filename
                
                (root, ext) = os.path.splitext(self.filename)
                
                actions = {'.tar.gz': self.import_targz,
                           '.tar.bz2': self.import_tarbz2,
                           '.xml': self.import_xml,
                           '.tar': self.import_tar}

                for k, v in actions.items():
                        if self.filename.endswith(k):
                                return v()
                        
                return _('File "%s" is not a valid snippets archive') % self.filename
# ex:ts=8:et:



# -*- coding: utf-8 -*-

# config.py -- Config dialog
#
# Copyright (C) 2008 - B. Clausius
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

# Parts from "Interactive Python-GTK Console" (stolen from epiphany's console.py)
#     Copyright (C), 1998 James Henstridge <james@daa.com.au>
#     Copyright (C), 2005 Adam Hooper <adamh@densi.com>
# Bits from pluma Python Console Plugin
#     Copyrignt (C), 2005 Raphaël Slinckx

import os
import gtk

__all__ = ('PythonConsoleConfig', 'PythonConsoleConfigDialog')

MATECONF_KEY_BASE = '/apps/pluma/plugins/pythonconsole'
MATECONF_KEY_COMMAND_COLOR = MATECONF_KEY_BASE + '/command-color'
MATECONF_KEY_ERROR_COLOR = MATECONF_KEY_BASE + '/error-color'

DEFAULT_COMMAND_COLOR = '#314e6c' # Blue Shadow
DEFAULT_ERROR_COLOR = '#990000' # Accent Red Dark

class PythonConsoleConfig(object):
    try:
        import mateconf
    except ImportError:
        mateconf = None

    def __init__(self):
        pass

    @staticmethod
    def enabled():
        return PythonConsoleConfig.mateconf != None

    @staticmethod
    def add_handler(handler):
        if PythonConsoleConfig.mateconf:
            PythonConsoleConfig.mateconf.client_get_default().notify_add(MATECONF_KEY_BASE, handler)

    color_command = property(
        lambda self: self.mateconf_get_str(MATECONF_KEY_COMMAND_COLOR, DEFAULT_COMMAND_COLOR),
        lambda self, value: self.mateconf_set_str(MATECONF_KEY_COMMAND_COLOR, value))

    color_error = property(
        lambda self: self.mateconf_get_str(MATECONF_KEY_ERROR_COLOR, DEFAULT_ERROR_COLOR),
        lambda self, value: self.mateconf_set_str(MATECONF_KEY_ERROR_COLOR, value))

    @staticmethod
    def mateconf_get_str(key, default=''):
        if not PythonConsoleConfig.mateconf:
            return default

        val = PythonConsoleConfig.mateconf.client_get_default().get(key)
        if val is not None and val.type == mateconf.VALUE_STRING:
            return val.get_string()
        else:
            return default

    @staticmethod
    def mateconf_set_str(key, value):
        if not PythonConsoleConfig.mateconf:
            return

        v = PythonConsoleConfig.mateconf.Value(mateconf.VALUE_STRING)
        v.set_string(value)
        PythonConsoleConfig.mateconf.client_get_default().set(key, v)

class PythonConsoleConfigDialog(object):

    def __init__(self, datadir):
        object.__init__(self)
        self._dialog = None
        self._ui_path = os.path.join(datadir, 'ui', 'config.ui')
        self.config = PythonConsoleConfig()

    def dialog(self):
        if self._dialog is None:
            self._ui = gtk.Builder()
            self._ui.add_from_file(self._ui_path)

            self.set_colorbutton_color(self._ui.get_object('colorbutton-command'),
                                        self.config.color_command)
            self.set_colorbutton_color(self._ui.get_object('colorbutton-error'),
                                        self.config.color_error)

            self._ui.connect_signals(self)

            self._dialog = self._ui.get_object('dialog-config')
            self._dialog.show_all()
        else:
            self._dialog.present()

        return self._dialog

    @staticmethod
    def set_colorbutton_color(colorbutton, value):
        try:
            color = gtk.gdk.color_parse(value)
        except ValueError:
            pass    # Default color in config.ui used
        else:
            colorbutton.set_color(color)

    def on_dialog_config_response(self, dialog, response_id):
        self._dialog.destroy()

    def on_dialog_config_destroy(self, dialog):
        self._dialog = None
        self._ui = None

    def on_colorbutton_command_color_set(self, colorbutton):
        self.config.color_command = colorbutton.get_color().to_string()

    def on_colorbutton_error_color_set(self, colorbutton):
        self.config.color_error = colorbutton.get_color().to_string()

# ex:et:ts=4:



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import re

import gtk
from gtk import gdk
import gio
import pluma
import gtksourceview2 as gsv
import gobject

from Library import Library
from Snippet import Snippet
from Placeholder import *
import Completion

class DynamicSnippet(dict):
        def __init__(self, text):
                self['text'] = text
                self.valid = True

class Document:
        TAB_KEY_VAL = (gtk.keysyms.Tab, \
                        gtk.keysyms.ISO_Left_Tab)
        SPACE_KEY_VAL = (gtk.keysyms.space,)
        
        def __init__(self, instance, view):
                self.view = None
                self.instance = instance
                
                self.placeholders = []
                self.active_snippets = []
                self.active_placeholder = None
                self.signal_ids = {}
                
                self.ordered_placeholders = []
                self.update_placeholders = []
                self.jump_placeholders = []
                self.language_id = 0
                self.timeout_update_id = 0
                
                self.provider = Completion.Provider(_('Snippets'), self.language_id, self.on_proposal_activated)
                self.defaults_provider = Completion.Defaults(self.on_default_activated)
                
                # Always have a reference to the global snippets
                Library().ref(None)
                self.set_view(view)
        
        # Stop controlling the view. Remove all active snippets, remove references
        # to the view and the plugin instance, disconnect all signal handlers
        def stop(self):
                if self.timeout_update_id != 0:
                        gobject.source_remove(self.timeout_update_id)
                        self.timeout_update_id = 0
                        del self.update_placeholders[:]
                        del self.jump_placeholders[:]

                # Always release the reference to the global snippets
                Library().unref(None)
                self.set_view(None)
                self.instance = None
                self.active_placeholder = None

        def disconnect_signal(self, obj, signal):
                if (obj, signal) in self.signal_ids:
                        obj.disconnect(self.signal_ids[(obj, signal)])
                        del self.signal_ids[(obj, signal)]
                
        def connect_signal(self, obj, signal, cb):
                self.disconnect_signal(obj, signal)     
                self.signal_ids[(obj, signal)] = obj.connect(signal, cb)

        def connect_signal_after(self, obj, signal, cb):
                self.disconnect_signal(obj, signal)
                self.signal_ids[(obj, signal)] = obj.connect_after(signal, cb)
                
        # Set the view to be controlled. Installs signal handlers and sets current
        # language. If there is already a view set this function will first remove
        # all currently active snippets and disconnect all current signals. So
        # self.set_view(None) will effectively remove all the control from the
        # current view
        def _set_view(self, view):
                if self.view:
                        buf = self.view.get_buffer()

                        # Remove signals
                        signals = {self.view: ('key-press-event', 'destroy', 
                                               'notify::editable', 'drag-data-received', 'expose-event'),
                                   buf:       ('notify::language', 'changed', 'cursor-moved', 'insert-text'),
                                   self.view.get_completion(): ('hide',)}

                        for obj, sig in signals.items():
                                for s in sig:
                                        self.disconnect_signal(obj, s)
                        
                        # Remove all active snippets
                        for snippet in list(self.active_snippets):
                                self.deactivate_snippet(snippet, True)
                        
                        completion = self.view.get_completion()
                        completion.remove_provider(self.provider)
                        completion.remove_provider(self.defaults_provider)                        

                self.view = view
                
                if view != None:
                        buf = view.get_buffer()
                       
                        self.connect_signal(view, 'destroy', self.on_view_destroy)

                        if view.get_editable():
                                self.connect_signal(view, 'key-press-event', self.on_view_key_press)

                        self.connect_signal(buf, 'notify::language', self.on_notify_language)
                        self.connect_signal(view, 'notify::editable', self.on_notify_editable)
                        self.connect_signal(view, 'drag-data-received', self.on_drag_data_received)
                        self.connect_signal_after(view, 'expose-event', self.on_expose_event)

                        self.update_language()

                        completion = view.get_completion()
                        completion.add_provider(self.provider)
                        
                        completion.add_provider(self.defaults_provider)
                        
                        self.connect_signal(completion, 'hide', self.on_completion_hide)
                elif self.language_id != 0:
                        langid = self.language_id
                        
                        self.language_id = None;
                        self.provider.language_id = self.language_id
                        
                        if self.instance:
                                self.instance.language_changed(self)

                        Library().unref(langid)

        def set_view(self, view):
                if view == self.view:
                        return
                
                self._set_view(view)

        # Call this whenever the language in the view changes. This makes sure that
        # the correct language is used when finding snippets
        def update_language(self):
                lang = self.view.get_buffer().get_language()

                if lang == None and self.language_id == None:
                        return
                elif lang and lang.get_id() == self.language_id:
                        return

                langid = self.language_id

                if lang:
                        self.language_id = lang.get_id()
                else:
                        self.language_id = None

                if self.instance:
                        self.instance.language_changed(self)

                if langid != 0:
                        Library().unref(langid)

                Library().ref(self.language_id)
                self.provider.language_id = self.language_id

        def accelerator_activate(self, keyval, mod):
                if not self.view or not self.view.get_editable():
                        return False

                accelerator = gtk.accelerator_name(keyval, mod)
                snippets = Library().from_accelerator(accelerator, \
                                self.language_id)

                snippets_debug('Accel!')

                if len(snippets) == 0:
                        return False
                elif len(snippets) == 1:
                        self.apply_snippet(snippets[0])
                else:
                        # Do the fancy completion dialog
                        self.provider.set_proposals(snippets)
                        self.view.show_completion((self,))

                return True

        def first_snippet_inserted(self):
                buf = self.view.get_buffer()
                
                self.connect_signal(buf, 'changed', self.on_buffer_changed)
                self.connect_signal(buf, 'cursor-moved', self.on_buffer_cursor_moved)
                self.connect_signal_after(buf, 'insert-text', self.on_buffer_insert_text)
        
        def last_snippet_removed(self):
                buf = self.view.get_buffer()
                self.disconnect_signal(buf, 'changed')
                self.disconnect_signal(buf, 'cursor-moved')
                self.disconnect_signal(buf, 'insert-text')

        def current_placeholder(self):
                buf = self.view.get_buffer()
                
                piter = buf.get_iter_at_mark(buf.get_insert())        
                found = []

                for placeholder in self.placeholders:
                        begin = placeholder.begin_iter()
                        end = placeholder.end_iter()

                        if piter.compare(begin) >= 0 and piter.compare(end) <= 0:
                                found.append(placeholder)

                if self.active_placeholder in found:
                        return self.active_placeholder
                elif len(found) > 0:
                        return found[0]
                else:
                        return None

        def advance_placeholder(self, direction):
                # Returns (CurrentPlaceholder, NextPlaceholder), depending on direction
                buf = self.view.get_buffer()
                
                piter = buf.get_iter_at_mark(buf.get_insert())
                found = current = next = None
                length = len(self.placeholders)
                
                placeholders = list(self.placeholders)
                
                if self.active_placeholder:
                        begin = self.active_placeholder.begin_iter()
                        end = self.active_placeholder.end_iter()
                        
                        if piter.compare(begin) >= 0 and piter.compare(end) <= 0:
                                current = self.active_placeholder
                                currentIndex = placeholders.index(self.active_placeholder)

                if direction == 1:
                        # w = piter, x = begin, y = end, z = found
                        nearest = lambda w, x, y, z: (w.compare(x) <= 0 and (not z or \
                                        x.compare(z.begin_iter()) < 0))
                        indexer = lambda x: x < length - 1
                else:
                        # w = piter, x = begin, y = end, z = prev
                        nearest = lambda w, x, y, z: (w.compare(x) >= 0 and (not z or \
                                        x.compare(z.begin_iter()) >= 0))
                        indexer = lambda x: x > 0

                for index in range(0, length):
                        placeholder = placeholders[index]
                        begin = placeholder.begin_iter()
                        end = placeholder.end_iter()
                        
                        # Find the nearest placeholder
                        if nearest(piter, begin, end, found):
                                foundIndex = index
                                found = placeholder
                        
                        # Find the current placeholder
                        if piter.compare(begin) >= 0 and \
                                        piter.compare(end) <= 0 and \
                                        current == None:
                                currentIndex = index
                                current = placeholder
                
                if current and current != found and \
                   (current.begin_iter().compare(found.begin_iter()) == 0 or \
                    current.end_iter().compare(found.begin_iter()) == 0) and \
                   self.active_placeholder and \
                   current.begin_iter().compare(self.active_placeholder.begin_iter()) == 0:
                        # if current and found are at the same place, then
                        # resolve the 'hugging' problem
                        current = self.active_placeholder
                        currentIndex = placeholders.index(current)
                
                if current:
                        if indexer(currentIndex):
                                next = placeholders[currentIndex + direction]
                elif found:
                        next = found
                elif length > 0:
                        next = self.placeholders[0]
                
                return current, next
        
        def next_placeholder(self):
                return self.advance_placeholder(1)
        
        def previous_placeholder(self):
                return self.advance_placeholder(-1)

        def cursor_on_screen(self):
                buf = self.view.get_buffer()
                self.view.scroll_mark_onscreen(buf.get_insert())
        
        def set_active_placeholder(self, placeholder):
                self.active_placeholder = placeholder

        def goto_placeholder(self, current, next):
                last = None

                if current:
                        # Signal this placeholder to end action
                        self.view.get_completion().hide()
                        current.leave()
                        
                        if current.__class__ == PlaceholderEnd:
                                last = current
                
                self.set_active_placeholder(next)
                
                if next:
                        next.enter()
                        
                        if next.__class__ == PlaceholderEnd:
                                last = next
                        elif len(next.defaults) > 1 and next.get_text() == next.default:
                                self.defaults_provider.set_defaults(next.defaults)
                                
                                cm = self.view.get_completion()
                                cm.show([self.defaults_provider], cm.create_context())

                if last:
                        # This is the end of the placeholder, remove the snippet etc
                        for snippet in list(self.active_snippets):
                                if snippet.placeholders[0] == last:
                                        self.deactivate_snippet(snippet)
                                        break
                
                self.cursor_on_screen()
                
                return next != None
        
        def skip_to_next_placeholder(self):
                (current, next) = self.next_placeholder()
                return self.goto_placeholder(current, next)
        
        def skip_to_previous_placeholder(self):
                (current, prev) = self.previous_placeholder()
                return self.goto_placeholder(current, prev)

        def env_get_selected_text(self, buf):
                bounds = buf.get_selection_bounds()

                if bounds:
                        return buf.get_text(bounds[0], bounds[1])
                else:
                        return ''

        def env_get_current_word(self, buf):
                start, end = buffer_word_boundary(buf)
                
                return buf.get_text(start, end)

        def env_get_current_line(self, buf):
                start, end = buffer_line_boundary(buf)
                
                return buf.get_text(start, end)

        def env_get_current_line_number(self, buf):
                start, end = buffer_line_boundary(buf)
                
                return str(start.get_line() + 1)
                
        def env_get_document_uri(self, buf):
                location = buf.get_location()
                
                if location:
                        return location.get_uri()
                else:
                        return ''
        
        def env_get_document_name(self, buf):
                location = buf.get_location()
                
                if location:
                        return location.get_basename()
                else:
                        return ''

        def env_get_document_scheme(self, buf):
                location = buf.get_location()
                
                if location:
                        return location.get_uri_scheme()
                else:
                        return ''

        def env_get_document_path(self, buf):
                location = buf.get_location()
                
                if location:
                        return location.get_path()
                else:
                        return ''

        def env_get_document_dir(self, buf):
                location = buf.get_location()

                if location:
                        return location.get_parent().get_path() or ''
                else:
                        return ''

        def env_get_document_type(self, buf):
                typ = buf.get_mime_type()
                
                if typ:
                        return typ
                else:
                        return ''

        def env_get_documents_uri(self, buf):
                toplevel = self.view.get_toplevel()
                
                if isinstance(toplevel, pluma.Window):
                        documents_uri = [doc.get_location().get_uri()
                                         for doc in toplevel.get_documents()
                                         if doc.get_location() is not None]
                else:
                        documents_uri = []
                
                return ' '.join(documents_uri)

        def env_get_documents_path(self, buf):
                toplevel = self.view.get_toplevel()
                
                if isinstance(toplevel, pluma.Window):
                        documents_location = [doc.get_location()
                                              for doc in toplevel.get_documents()
                                              if doc.get_location() is not None]

                        documents_path = [location.get_path()
                                          for location in documents_location
                                          if pluma.utils.uri_has_file_scheme(location.get_uri())]
                else:
                        documents_path = []
                
                return ' '.join(documents_path)

        def update_environment(self):
                buf = self.view.get_buffer()
                
                variables = {'PLUMA_SELECTED_TEXT': self.env_get_selected_text, 
                             'PLUMA_CURRENT_WORD': self.env_get_current_word, 
                             'PLUMA_CURRENT_LINE': self.env_get_current_line,
                             'PLUMA_CURRENT_LINE_NUMBER': self.env_get_current_line_number,
                             'PLUMA_CURRENT_DOCUMENT_URI': self.env_get_document_uri, 
                             'PLUMA_CURRENT_DOCUMENT_NAME': self.env_get_document_name,
                             'PLUMA_CURRENT_DOCUMENT_SCHEME': self.env_get_document_scheme,
                             'PLUMA_CURRENT_DOCUMENT_PATH': self.env_get_document_path,
                             'PLUMA_CURRENT_DOCUMENT_DIR': self.env_get_document_dir,
                             'PLUMA_CURRENT_DOCUMENT_TYPE': self.env_get_document_type,
                             'PLUMA_DOCUMENTS_URI': self.env_get_documents_uri,
                             'PLUMA_DOCUMENTS_PATH': self.env_get_documents_path,
                             }
                
                for var in variables:
                        os.environ[var] = variables[var](buf)
        
        def uses_current_word(self, snippet):
                matches = re.findall('(\\\\*)\\$PLUMA_CURRENT_WORD', snippet['text'])
                
                for match in matches:
                        if len(match) % 2 == 0:
                                return True
                
                return False
        
        def uses_current_line(self, snippet):
                matches = re.findall('(\\\\*)\\$PLUMA_CURRENT_LINE', snippet['text'])
                
                for match in matches:
                        if len(match) % 2 == 0:
                                return True
                
                return False

        def apply_snippet(self, snippet, start = None, end = None):
                if not snippet.valid:
                        return False

                buf = self.view.get_buffer()
                s = Snippet(snippet)
                
                if not start:
                        start = buf.get_iter_at_mark(buf.get_insert())
                
                if not end:
                        end = buf.get_iter_at_mark(buf.get_selection_bound())

                if start.equal(end) and self.uses_current_word(s):
                        # There is no tab trigger and no selection and the snippet uses
                        # the current word. Set start and end to the word boundary so that 
                        # it will be removed
                        start, end = buffer_word_boundary(buf)
                elif start.equal(end) and self.uses_current_line(s):
                        # There is no tab trigger and no selection and the snippet uses
                        # the current line. Set start and end to the line boundary so that 
                        # it will be removed
                        start, end = buffer_line_boundary(buf)

                # Set environmental variables
                self.update_environment()
                
                # You know, we could be in an end placeholder
                (current, next) = self.next_placeholder()
                if current and current.__class__ == PlaceholderEnd:
                        self.goto_placeholder(current, None)
                
                buf.begin_user_action()

                # Remove the tag, selection or current word
                buf.delete(start, end)
                
                # Insert the snippet
                holders = len(self.placeholders)
                
                if len(self.active_snippets) == 0:
                        self.first_snippet_inserted()

                sn = s.insert_into(self, start)
                self.active_snippets.append(sn)

                # Put cursor at first tab placeholder
                keys = filter(lambda x: x > 0, sn.placeholders.keys())

                if len(keys) == 0:
                        if 0 in sn.placeholders:
                                self.goto_placeholder(self.active_placeholder, sn.placeholders[0])
                        else:
                                buf.place_cursor(sn.begin_iter())
                else:
                        self.goto_placeholder(self.active_placeholder, sn.placeholders[keys[0]])

                if sn in self.active_snippets:
                        # Check if we can get end_iter in view without moving the
                        # current cursor position out of view
                        cur = buf.get_iter_at_mark(buf.get_insert())
                        last = sn.end_iter()

                        curloc = self.view.get_iter_location(cur)
                        lastloc = self.view.get_iter_location(last)

                        if (lastloc.y + lastloc.height) - curloc.y <= \
                           self.view.get_visible_rect().height:
                                self.view.scroll_mark_onscreen(sn.end_mark)

                buf.end_user_action()
                self.view.grab_focus()

                return True

        def get_tab_tag(self, buf, end = None):
                if not end:
                        end = buf.get_iter_at_mark(buf.get_insert())

                start = end.copy()
                
                word = None
                
                if start.backward_word_start():
                        # Check if we were at a word start ourselves
                        tmp = start.copy()
                        tmp.forward_word_end()
                        
                        if tmp.equal(end):
                                word = buf.get_text(start, end)
                        else:
                                start = end.copy()
                else:
                        start = end.copy()
                
                if not word or word == '':
                        if start.backward_char():
                                word = start.get_char()

                                if word.isalnum() or word.isspace():
                                        return (None, None, None)
                        else:
                                return (None, None, None)

                return (word, start, end)

        def parse_and_run_snippet(self, data, iter):
                self.apply_snippet(DynamicSnippet(data), iter, iter)

        def run_snippet_trigger(self, trigger, bounds):
                if not self.view:
                        return False

                snippets = Library().from_tag(trigger, self.language_id)
                buf = self.view.get_buffer()

                if snippets:
                        if len(snippets) == 1:
                                return self.apply_snippet(snippets[0], bounds[0], bounds[1])
                        else:
                                # Do the fancy completion dialog
                                self.provider.set_proposals(snippets)
                                cm = self.view.get_completion()
                                
                                cm.show([self.provider], cm.create_context())
                                return True

                return False
                
        def run_snippet(self):
                if not self.view:
                        return False

                buf = self.view.get_buffer()

                # get the word preceding the current insertion position
                (word, start, end) = self.get_tab_tag(buf)

                if not word:
                        return self.skip_to_next_placeholder()

                if not self.run_snippet_trigger(word, (start, end)):
                        return self.skip_to_next_placeholder()
                else:
                        return True

        def deactivate_snippet(self, snippet, force = False):
                buf = self.view.get_buffer()
                remove = []
                ordered_remove = []

                for tabstop in snippet.placeholders:
                        if tabstop == -1:
                                placeholders = snippet.placeholders[-1]
                        else:
                                placeholders = [snippet.placeholders[tabstop]]
                        
                        for placeholder in placeholders:
                                if placeholder in self.placeholders:
                                        if placeholder in self.update_placeholders:
                                                placeholder.update_contents()
                                                
                                                self.update_placeholders.remove(placeholder)
                                        elif placeholder in self.jump_placeholders:
                                                placeholder[0].leave()

                                        remove.append(placeholder)
                                elif placeholder in self.ordered_placeholders:
                                        ordered_remove.append(placeholder)

                for placeholder in remove:
                        if placeholder == self.active_placeholder:
                                self.active_placeholder = None

                        self.placeholders.remove(placeholder)
                        self.ordered_placeholders.remove(placeholder)

                        placeholder.remove(force)

                for placeholder in ordered_remove:
                        self.ordered_placeholders.remove(placeholder)
                        placeholder.remove(force)

                snippet.deactivate()
                self.active_snippets.remove(snippet)

                if len(self.active_snippets) == 0:
                        self.last_snippet_removed()

                self.view.queue_draw()

        def update_snippet_contents(self):
                self.timeout_update_id = 0

                for placeholder in self.update_placeholders:
                        placeholder.update_contents()

                for placeholder in self.jump_placeholders:
                        self.goto_placeholder(placeholder[0], placeholder[1])

                del self.update_placeholders[:]
                del self.jump_placeholders[:]

                return False

        # Callbacks
        def on_view_destroy(self, view):
                self.stop()
                return

        def on_buffer_cursor_moved(self, buf):
                piter = buf.get_iter_at_mark(buf.get_insert())

                # Check for all snippets if the cursor is outside its scope
                for snippet in list(self.active_snippets):
                        if snippet.begin_mark.get_deleted() or snippet.end_mark.get_deleted():
                                self.deactivate(snippet)
                        else:
                                begin = snippet.begin_iter()
                                end = snippet.end_iter()
                        
                                if piter.compare(begin) < 0 or piter.compare(end) > 0:
                                        # Oh no! Remove the snippet this instant!!
                                        self.deactivate_snippet(snippet)

                current = self.current_placeholder()

                if current != self.active_placeholder:
                        self.jump_placeholders.append((self.active_placeholder, current))

                        if self.timeout_update_id == 0:
                                self.timeout_update_id = gobject.timeout_add(0, 
                                                self.update_snippet_contents)

        def on_buffer_changed(self, buf):
                current = self.current_placeholder()

                if current:
                        if not current in self.update_placeholders:
                                self.update_placeholders.append(current)
                
                        if self.timeout_update_id == 0:
                                self.timeout_update_id = gobject.timeout_add(0, \
                                                self.update_snippet_contents)
        
        def on_buffer_insert_text(self, buf, piter, text, length):
                ctx = get_buffer_context(buf)
                
                # do nothing special if there is no context and no active 
                # placeholder
                if (not ctx) and (not self.active_placeholder):
                        return
                
                if not ctx:
                        ctx = self.active_placeholder
                
                if not ctx in self.ordered_placeholders:
                        return
                        
                # move any marks that were incorrectly moved by this insertion
                # back to where they belong
                begin = ctx.begin_iter()
                end = ctx.end_iter()
                idx = self.ordered_placeholders.index(ctx)
                
                for placeholder in self.ordered_placeholders:
                        if placeholder == ctx:
                                continue
                        
                        ob = placeholder.begin_iter()
                        oe = placeholder.end_iter()
                        
                        if ob.compare(begin) == 0 and ((not oe) or oe.compare(end) == 0):
                                oidx = self.ordered_placeholders.index(placeholder)
                                
                                if oidx > idx and ob:
                                        buf.move_mark(placeholder.begin, end)
                                elif oidx < idx and oe:
                                        buf.move_mark(placeholder.end, begin)
                        elif ob.compare(begin) >= 0 and ob.compare(end) < 0 and (oe and oe.compare(end) >= 0):
                                buf.move_mark(placeholder.begin, end)
                        elif (oe and oe.compare(begin) > 0) and ob.compare(begin) <= 0:
                                buf.move_mark(placeholder.end, begin)
        
        def on_notify_language(self, buf, spec):
                self.update_language()

        def on_notify_editable(self, view, spec):
                self._set_view(view)
        
        def on_view_key_press(self, view, event):
                library = Library()

                if not (event.state & gdk.CONTROL_MASK) and \
                                not (event.state & gdk.MOD1_MASK) and \
                                event.keyval in self.TAB_KEY_VAL:
                        if not event.state & gdk.SHIFT_MASK:
                                return self.run_snippet()
                        else:
                                return self.skip_to_previous_placeholder()
                elif not library.loaded and \
                                library.valid_accelerator(event.keyval, event.state):
                        library.ensure_files()
                        library.ensure(self.language_id)
                        self.accelerator_activate(event.keyval, \
                                        event.state & gtk.accelerator_get_default_mod_mask())

                return False
        
        def path_split(self, path, components=[]):
                head, tail = os.path.split(path)
                
                if not tail and head:
                        return [head] + components
                elif tail:
                        return self.path_split(head, [tail] + components)
                else:
                        return components

        def relative_path(self, first, second, mime):
                prot1 = re.match('(^[a-z]+:\/\/|\/)(.*)', first)
                prot2 = re.match('(^[a-z]+:\/\/|\/)(.*)', second)

                if not prot1 or not prot2:
                        return second

                # Different protocols
                if prot1.group(1) != prot2.group(1):
                        return second

                # Split on backslash
                path1 = self.path_split(prot1.group(2))
                path2 = self.path_split(prot2.group(2))

                # Remove as long as common
                while path1 and path2 and path1[0] == path2[0]:
                        path1.pop(0)
                        path2.pop(0)

                # If we need to ../ more than 3 times, then just return
                # the absolute path
                if len(path1) - 1 > 3:
                        return second

                if mime.startswith('x-directory'):
                        # directory, special case
                        if not path2:
                                result = './'
                        else:
                                result = '../' * (len(path1) - 1)
                else:
                        # Insert ../
                        result = '../' * (len(path1) - 1)

                        if not path2:
                                result = os.path.basename(second)
                
                if path2:
                        result += os.path.join(*path2)

                return result

        def apply_uri_snippet(self, snippet, mime, uri):
                # Remove file scheme
                gfile = gio.File(uri)
                pathname = ''
                dirname = ''
                ruri = ''

                if pluma.utils.uri_has_file_scheme(uri):
                        pathname = gfile.get_path()
                        dirname = gfile.get_parent().get_path()

                name = os.path.basename(uri)
                scheme = gfile.get_uri_scheme()

                os.environ['PLUMA_DROP_DOCUMENT_URI'] = uri
                os.environ['PLUMA_DROP_DOCUMENT_NAME'] = name
                os.environ['PLUMA_DROP_DOCUMENT_SCHEME'] = scheme
                os.environ['PLUMA_DROP_DOCUMENT_PATH'] = pathname
                os.environ['PLUMA_DROP_DOCUMENT_DIR'] = dirname
                os.environ['PLUMA_DROP_DOCUMENT_TYPE'] = mime

                buf = self.view.get_buffer()
                location = buf.get_location()
                if location:
                        ruri = location.get_uri()

                relpath = self.relative_path(ruri, uri, mime)

                os.environ['PLUMA_DROP_DOCUMENT_RELATIVE_PATH'] = relpath

                mark = buf.get_mark('gtk_drag_target')
                
                if not mark:
                        mark = buf.get_insert()

                piter = buf.get_iter_at_mark(mark)
                self.apply_snippet(snippet, piter, piter)

        def in_bounds(self, x, y):
                rect = self.view.get_visible_rect()
                rect.x, rect.y = self.view.buffer_to_window_coords(gtk.TEXT_WINDOW_WIDGET, rect.x, rect.y)

                return not (x < rect.x or x > rect.x + rect.width or y < rect.y or y > rect.y + rect.height)
        
        def on_drag_data_received(self, view, context, x, y, data, info, timestamp):   
                if not (gtk.targets_include_uri(context.targets) and data.data and self.in_bounds(x, y)):
                        return

                uris = drop_get_uris(data)
                uris.reverse()
                stop = False
                
                for uri in uris:
                        try:
                                mime = gio.content_type_guess(uri)
                        except:
                                mime = None

                        if not mime:
                                continue
                        
                        snippets = Library().from_drop_target(mime, self.language_id)
                        
                        if snippets:
                                stop = True
                                self.apply_uri_snippet(snippets[0], mime, uri)

                if stop:
                        context.finish(True, False, timestamp)
                        view.stop_emission('drag-data-received')
                        view.get_toplevel().present()
                        view.grab_focus()
        
        def find_uri_target(self, context):
                lst = gtk.target_list_add_uri_targets((), 0)
                
                return self.view.drag_dest_find_target(context, lst)
        
        def on_completion_hide(self, completion):
                self.provider.set_proposals(None)

        def on_proposal_activated(self, proposal, piter):
                buf = self.view.get_buffer()
                bounds = buf.get_selection_bounds()
                
                if bounds:
                        self.apply_snippet(proposal.snippet(), None, None)
                else:
                        (word, start, end) = self.get_tab_tag(buf, piter)
                        self.apply_snippet(proposal.snippet(), start, end)

                return True
        
        def on_default_activated(self, proposal, piter):
                buf = self.view.get_buffer()
                bounds = buf.get_selection_bounds()

                if bounds:
                        buf.begin_user_action()
                        buf.delete(bounds[0], bounds[1])
                        buf.insert(bounds[0], proposal.props.label)
                        buf.end_user_action()

                        return True
                else:
                        return False

        def iter_coords(self, piter):
                rect = self.view.get_iter_location(piter)
                rect.x, rect.y = self.view.buffer_to_window_coords(gtk.TEXT_WINDOW_TEXT, rect.x, rect.y)

                return rect

        def placeholder_in_area(self, placeholder, area):
                start = placeholder.begin_iter()
                end = placeholder.end_iter()

                if not start or not end:
                        return False

                # Test if start is before bottom, and end is after top
                start_rect = self.iter_coords(start)
                end_rect = self.iter_coords(end)

                return start_rect.y <= area.y + area.height and \
                       end_rect.y + end_rect.height >= area.y

        def draw_placeholder_rect(self, ctx, placeholder, col):
                start = placeholder.begin_iter()
                start_rect = self.iter_coords(start)
                start_line = start.get_line()

                end = placeholder.end_iter()
                end_rect = self.iter_coords(end)
                end_line = end.get_line()

                line = start.copy()
                line.set_line_offset(0)
                geom = self.view.get_window(gtk.TEXT_WINDOW_TEXT).get_geometry()
                
                ctx.translate(0.5, 0.5)
                
                while line.get_line() <= end_line:
                        ypos, height = self.view.get_line_yrange(line)
                        x_, ypos = self.view.window_to_buffer_coords(gtk.TEXT_WINDOW_TEXT, 0, ypos)

                        if line.get_line() == start_line and line.get_line() == end_line:
                                # Simply draw a box, both are on the same line
                                ctx.rectangle(start_rect.x, start_rect.y, end_rect.x - start_rect.x, start_rect.height - 1)
                                ctx.stroke()
                        elif line.get_line() == start_line or line.get_line() == end_line:
                                if line.get_line() == start_line:
                                        rect = start_rect
                                else:
                                        rect = end_rect

                                ctx.move_to(0, rect.y + rect.height - 1)
                                ctx.rel_line_to(rect.x, 0)
                                ctx.rel_line_to(0, -rect.height + 1)
                                ctx.rel_line_to(geom[2], 0)
                                ctx.stroke()

                        if not line.forward_line():
                                break

        def draw_placeholder_bar(self, ctx, placeholder, col):
                start = placeholder.begin_iter()
                start_rect = self.iter_coords(start)
                
                ctx.translate(0.5, 0.5)
                extend_width = 2.5

                ctx.move_to(start_rect.x - extend_width, start_rect.y)
                ctx.rel_line_to(extend_width * 2, 0)

                ctx.move_to(start_rect.x, start_rect.y)
                ctx.rel_line_to(0, start_rect.height - 1)
                
                ctx.rel_move_to(-extend_width, 0)
                ctx.rel_line_to(extend_width * 2, 0)
                ctx.stroke()

        def from_color(self, col):
                return [col.red / 0x10000, col.green / 0x10000, col.blue / 0x10000]

        def draw_placeholder(self, ctx, placeholder):
                if isinstance(placeholder, PlaceholderEnd):
                        return

                buf = self.view.get_buffer()

                col = self.from_color(self.view.get_style().text[gtk.STATE_INSENSITIVE])
                ctx.set_source_rgba(col[0], col[1], col[2], 0.5)
                
                if placeholder.tabstop > 0:
                        ctx.set_dash([], 0)
                else:
                        ctx.set_dash([2], 0)

                start = placeholder.begin_iter()
                end = placeholder.end_iter()

                if start.equal(end):
                        self.draw_placeholder_bar(ctx, placeholder, col)
                else:
                        self.draw_placeholder_rect(ctx, placeholder, col)

        def on_expose_event(self, view, event):
                if event.window != view.get_window(gtk.TEXT_WINDOW_TEXT):
                        return False

                # Draw something
                ctx = event.window.cairo_create()
                ctx.rectangle(event.area)
                ctx.clip()

                ctx.set_line_width(1.0)

                for placeholder in self.ordered_placeholders:
                        if not self.placeholder_in_area(placeholder, event.area):
                                continue

                        ctx.save()
                        self.draw_placeholder(ctx, placeholder)
                        ctx.restore()

                return False

# ex:ts=8:et:



import gtksourceview2 as gsv
import gobject
import pluma
import gtk

from Library import Library
from LanguageManager import get_language_manager
from Snippet import Snippet

class Proposal(gobject.GObject, gsv.CompletionProposal):
        def __init__(self, snippet):
                gobject.GObject.__init__(self)
                self._snippet = Snippet(snippet)
        
        def snippet(self):
                return self._snippet.data
        
        # Interface implementation
        def do_get_markup(self):
                return self._snippet.display()
        
        def do_get_info(self):
                return self._snippet.data['text']

class Provider(gobject.GObject, gsv.CompletionProvider):
        def __init__(self, name, language_id, handler):
                gobject.GObject.__init__(self)
                
                self.name = name
                self.info_widget = None
                self.proposals = []
                self.language_id = language_id
                self.handler = handler
                self.info_widget = None
                self.mark = None
                
                theme = gtk.icon_theme_get_default()
                w, h = gtk.icon_size_lookup(gtk.ICON_SIZE_MENU)

                self.icon = theme.load_icon(gtk.STOCK_JUSTIFY_LEFT, w, 0)
        
        def __del__(self):
                if self.mark:
                        self.mark.get_buffer().delete_mark(self.mark)
        
        def set_proposals(self, proposals):
                self.proposals = proposals

        def mark_position(self, it):
                if not self.mark:
                        self.mark = it.get_buffer().create_mark(None, it, True)
                else:
                        self.mark.get_buffer().move_mark(self.mark, it)
        
        def get_word(self, context):
                it = context.get_iter()
                
                if it.starts_word() or it.starts_line() or not it.ends_word():
                        return None
                
                start = it.copy()
                
                if start.backward_word_start():
                        self.mark_position(start)
                        return start.get_text(it)
                else:
                        return None
        
        def do_get_start_iter(self, context, proposal):
                if not self.mark or self.mark.get_deleted():
                        return None
                
                return self.mark.get_buffer().get_iter_at_mark(self.mark)
                
        def do_match(self, context):
                return True

        def get_proposals(self, word):
                if self.proposals:
                        proposals = self.proposals
                else:
                        proposals = Library().get_snippets(None)
                        
                        if self.language_id:
                                proposals += Library().get_snippets(self.language_id)

                # Filter based on the current word
                if word:
                        proposals = filter(lambda x: x['tag'].startswith(word), proposals)

                return map(lambda x: Proposal(x), proposals)

        def do_populate(self, context):
                proposals = self.get_proposals(self.get_word(context))
                context.add_proposals(self, proposals, True)

        def do_get_name(self):
                return self.name

        def do_activate_proposal(self, proposal, piter):
                return self.handler(proposal, piter)
        
        def do_get_info_widget(self, proposal):
                if not self.info_widget:
                        view = pluma.View(pluma.Document())
                        manager = get_language_manager()

                        lang = manager.get_language('snippets')
                        view.get_buffer().set_language(lang)
                        
                        sw = gtk.ScrolledWindow()
                        sw.add(view)
                        
                        self.info_view = view
                        self.info_widget = sw
                
                return self.info_widget
        
        def do_update_info(self, proposal, info):
                buf = self.info_view.get_buffer()
                
                buf.set_text(proposal.get_info())
                buf.move_mark(buf.get_insert(), buf.get_start_iter())
                buf.move_mark(buf.get_selection_bound(), buf.get_start_iter())
                self.info_view.scroll_to_iter(buf.get_start_iter(), False)

                info.set_sizing(-1, -1, False, False)
                info.process_resize()
        
        def do_get_icon(self):
                return self.icon

        def do_get_activation(self):
                return gsv.COMPLETION_ACTIVATION_USER_REQUESTED

class Defaults(gobject.GObject, gsv.CompletionProvider):
        def __init__(self, handler):
                gobject.GObject.__init__(self)

                self.handler = handler
                self.proposals = []
        
        def set_defaults(self, defaults):
                self.proposals = []
                
                for d in defaults:
                        self.proposals.append(gsv.CompletionItem(d))
                
        def do_get_name(self):
                return ""
        
        def do_activate_proposal(self, proposal, piter):
                return self.handler(proposal, piter)
        
        def do_populate(self, context):
                context.add_proposals(self, self.proposals, True)

        def do_get_activation(self):
                return gsv.COMPLETION_ACTIVATION_NONE

gobject.type_register(Proposal)
gobject.type_register(Provider)
gobject.type_register(Defaults)

# ex:ts=8:et:



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import re
import os
import gettext

import gtk
from gtk import gdk
import pluma

from Document import Document
from Library import Library

class WindowHelper:
        def __init__(self, plugin):
                self.plugin = plugin
                self.current_controller = None
                self.current_language = None
                self.signal_ids = {}
                
        def run(self, window):
                self.window = window

                self.insert_menu()
                self.register_messages()
                
                self.accel_group = Library().get_accel_group(None)
                
                window.add_accel_group(self.accel_group)
                window.connect('tab-added', self.on_tab_added)
                
                # Add controllers to all the current views
                for view in self.window.get_views():
                        if isinstance(view, pluma.View) and not self.has_controller(view):
                                view._snippet_controller = Document(self, view)
                
                self.update()
        
        def stop(self):
                self.window.remove_accel_group(self.accel_group)
                self.accel_group = None
                
                #self.window.remove_accel_group(accel)
                self.remove_menu()
                self.unregister_messages()

                # Iterate over all the tabs and remove every controller
                for view in self.window.get_views():
                        if isinstance(view, pluma.View) and self.has_controller(view):
                                view._snippet_controller.stop()
                                view._snippet_controller = None
                
                self.window = None
                self.plugin = None
        
        def register_messages(self):
                bus = self.window.get_message_bus()
                
                self.messages = {
                        'activate': bus.register('/plugins/snippets', 'activate', ('view', 'iter'), trigger=str, view=pluma.View, iter=gtk.TextIter),
                        'parse-and-activate': bus.register('/plugins/snippets', 'parse-and-activate', ('view', 'iter'), snippet=str, view=pluma.View, iter=gtk.TextIter)
                }
                
                bus.connect('/plugins/snippets', 'activate', self.on_message_activate)
                bus.connect('/plugins/snippets', 'parse-and-activate', self.on_message_parse_and_activate)
        
        def unregister_messages(self):
                bus = self.window.get_message_bus()
                
                for name in self.messages:
                        bus.unregister(self.messages[name])
                
                self.messages = {}
        
        def on_message_activate(self, bus, message):
                if message.has_key('view'):
                        view = message.view
                else:
                        view = self.window.get_active_view()
                
                if not self.has_controller(view):
                        return
                
                if message.has_key('iter'):
                        iter = message.iter
                else:
                        iter = view.get_buffer().get_iter_at_mark(view.get_buffer().get_insert())
                
                controller = view._snippet_controller
                controller.run_snippet_trigger(message.trigger, (iter, iter))

        def on_message_parse_and_activate(self, bus, message):
                if message.has_key('view'):
                        view = message.view
                else:
                        view = self.window.get_active_view()
                
                if not self.has_controller(view):
                        return
                
                if message.has_key('iter'):
                        iter = message.iter
                else:
                        iter = view.get_buffer().get_iter_at_mark(view.get_buffer().get_insert())
                
                controller = view._snippet_controller
                controller.parse_and_run_snippet(message.snippet, iter)

        def insert_menu(self):
                manager = self.window.get_ui_manager()

                self.action_group = gtk.ActionGroup("PlumaSnippetPluginActions")
                self.action_group.set_translation_domain('pluma')
                self.action_group.add_actions([('ManageSnippets', None,
                                _('Manage _Snippets...'), \
                                None, _('Manage snippets'), \
                                self.on_action_snippets_activate)])

                self.merge_id = manager.new_merge_id()
                manager.insert_action_group(self.action_group, -1)
                manager.add_ui(self.merge_id, '/MenuBar/ToolsMenu/ToolsOps_5', \
                                'ManageSnippets', 'ManageSnippets', gtk.UI_MANAGER_MENUITEM, False)

        def remove_menu(self):
                manager = self.window.get_ui_manager()
                manager.remove_ui(self.merge_id)
                manager.remove_action_group(self.action_group)
                self.action_group = None
        
        def find_snippet(self, snippets, tag):
                result = []
                
                for snippet in snippets:
                        if Snippet(snippet)['tag'] == tag:
                                result.append(snippet)
                
                return result

        def has_controller(self, view):
                return hasattr(view, '_snippet_controller') and view._snippet_controller

        def update_language(self):
                if not self.window:
                        return

                if self.current_language:
                        accel_group = Library().get_accel_group( \
                                        self.current_language)
                        self.window.remove_accel_group(accel_group)

                if self.current_controller:
                        self.current_language = self.current_controller.language_id
                        
                        if self.current_language != None:
                                accel_group = Library().get_accel_group( \
                                                self.current_language)
                                self.window.add_accel_group(accel_group)
                else:
                        self.current_language = None
                
        def language_changed(self, controller):
                if controller == self.current_controller:
                        self.update_language()
                
        def update(self):
                view = self.window.get_active_view()
                
                if not view or not self.has_controller(view):
                        return
                
                controller = view._snippet_controller
                
                if controller != self.current_controller:
                        self.current_controller = controller
                        self.update_language()

        # Callbacks
        
        def on_tab_added(self, window, tab):
                # Create a new controller for this tab if it has a standard pluma view
                view = tab.get_view()
                
                if isinstance(view, pluma.View) and not self.has_controller(view):
                        view._snippet_controller = Document(self, view)

                self.update()

        def on_action_snippets_activate(self, item):
                self.plugin.create_configure_dialog()

        def accelerator_activated(self, keyval, mod):
                return self.current_controller.accelerator_activate(keyval, mod)

# ex:ts=8:et:
